CALL_TYPE_GET_LOCAL_COROSYNC_CONF = "CALL_TYPE_GET_LOCAL_COROSYNC_CONF"

class Call:
    type = CALL_TYPE_GET_LOCAL_COROSYNC_CONF

    def __init__(self, content):
        self.content = content

    def __repr__(self):
        return str("<GetLocalCorosyncConf>")


def get_get_local_corosync_conf(call_queue):
    def get_local_corosync_conf():
        _, expected_call = call_queue.take(CALL_TYPE_GET_LOCAL_COROSYNC_CONF)
        return expected_call.content
    return get_local_corosync_conf



import os.path
import re

from pcs import settings
from pcs.lib import reports
from pcs.lib.errors import LibraryError

def get_local_corosync_conf():
    """
    Read corosync.conf file from local machine
    """
    path = settings.corosync_conf_file
    try:
        with open(path, "r", encoding="utf-8") as a_file:
            return a_file.read()
    except EnvironmentError as e:
        raise LibraryError(reports.corosync_config_read_error(path, e.strerror))

def get_quorum_status_text(runner):
    """
    Get runtime quorum status from the local node
    """
    stdout, stderr, retval = runner.run([
        os.path.join(settings.corosync_binaries, "corosync-quorumtool"),
        "-p"
    ])
    # retval is 0 on success if the node is not in a partition with quorum
    # retval is 1 on error OR on success if the node has quorum
    if retval not in [0, 1] or stderr.strip():
        raise QuorumStatusReadException(stderr)
    return stdout

def set_expected_votes(runner, votes):
    """
    set expected votes in live cluster to specified value
    """
    stdout, stderr, retval = runner.run([
        os.path.join(settings.corosync_binaries, "corosync-quorumtool"),
        # format votes to handle the case where they are int
        "-e", "{0}".format(votes)
    ])
    if retval != 0:
        raise LibraryError(
            reports.corosync_quorum_set_expected_votes_error(stderr)
        )
    return stdout


class QuorumStatusException(Exception):
    def __init__(self, reason=""):
        super().__init__()
        self.reason = reason


class QuorumStatusReadException(QuorumStatusException):
    pass


class QuorumStatusParsingException(QuorumStatusException):
    pass


class QuorumStatus():
    """
    Parse corosync quourm status and provide access to parsed info
    """
    def __init__(self, data):
        self._data = data

    @classmethod
    def from_string(cls, quorum_status):
        # pylint: disable=too-many-branches
        """
        Create an instance based on corosync quorum status plaintext data

        string quorum_status -- corosync quorum status plaintext
        """
        parsed = {}
        in_node_list = False
        try:
            for line in quorum_status.splitlines():
                line = line.strip()
                if not line:
                    continue
                if in_node_list:
                    if line.startswith("-") or line.startswith("Nodeid"):
                        # skip headers
                        continue
                    parts = line.split()
                    if parts[0] == "0":
                        # this line has nodeid == 0, this is a qdevice line
                        parsed["qdevice_list"].append({
                            "name": parts[2],
                            "votes": int(parts[1]),
                            "local": False,
                        })
                    else:
                        # this line has non-zero nodeid, this is a node line
                        parsed["node_list"].append({
                            "name": parts[3],
                            "votes": int(parts[1]),
                            "local": len(parts) > 4 and parts[4] == "(local)",
                        })
                else:
                    if line == "Membership information":
                        in_node_list = True
                        parsed["node_list"] = []
                        parsed["qdevice_list"] = []
                        continue
                    if not ":" in line:
                        continue
                    parts = [x.strip() for x in line.split(":", 1)]
                    if parts[0] == "Quorate":
                        parsed["quorate"] = parts[1].lower() == "yes"
                    elif parts[0] == "Quorum":
                        match = re.match(r"(\d+).*", parts[1])
                        if match:
                            parsed["quorum"] = int(match.group(1))
                        else:
                            raise QuorumStatusParsingException(
                                "Unable to read number of votes needed for "
                                "quorum"
                            )
        except (ValueError, IndexError):
            raise QuorumStatusParsingException()
        missing_sections = [
            required
            for required in ("quorum", "quorate", "node_list")
            if required not in parsed
        ]
        if missing_sections:
            raise QuorumStatusParsingException(
                "Missing required section(s): '{}'".format(
                    "', '".join(sorted(missing_sections))
                )
            )
        return cls(parsed)

    @property
    def is_quorate(self):
        """
        Does the cluster have quorum?
        """
        return bool(self._data["quorate"])

    @property
    def votes_needed_for_quorum(self):
        """
        How many votes are needed for the cluster to have quorum?
        """
        return self._data["quorum"]

    @property
    def qdevice_votes(self):
        """
        How many votes are provided by qdevice(s)?
        """
        qdevice_votes = 0
        for qdevice_info in self._data.get("qdevice_list", []):
            qdevice_votes += qdevice_info["votes"]
        return qdevice_votes

    def _get_votes_excluding_nodes(self, node_names):
        """
        How many votes do remain if some nodes are not counted in?

        iterable node_names -- list of names of nodes not to count in
        """
        votes = 0
        for node_info in self._data["node_list"]:
            if not (node_names and node_info["name"] in node_names):
                votes += node_info["votes"]
        votes += self.qdevice_votes
        return votes

    def _get_votes_excluding_local_node(self):
        """
        How many votes do remain if the local node is not counted in?
        """
        votes = 0
        for node_info in self._data["node_list"]:
            if not node_info["local"]:
                votes += node_info["votes"]
        votes += self.qdevice_votes
        return votes

    def stopping_nodes_cause_quorum_loss(self, node_names):
        """
        Will quorum be lost if specified nodes are stopped?

        iterable node_names -- list of names of nodes to be stopped
        """
        if not self.is_quorate:
            return False
        return (
            self._get_votes_excluding_nodes(node_names)
            <
            self.votes_needed_for_quorum
        )

    def stopping_local_node_cause_quorum_loss(self):
        """
        Will quorum be lost if the local node is stopped?
        """
        if not self.is_quorate:
            return False
        return (
            self._get_votes_excluding_local_node()
            <
            self.votes_needed_for_quorum
        )



import logging

LOGGER_NAMES = [
    "pcs.daemon",
    "tornado.application",
    "tornado.access",
    "tornado.general",
]

#pylint:disable=invalid-name
pcsd = logging.getLogger("pcs.daemon")

def from_external_source(level, created: float, usecs: int, message, group_id):
    record = pcsd.makeRecord(
        name=pcsd.name,
        level=level,
        # Information about stack fram is not needed here. Values are
        # inspired by the code of the logging module.
        fn="(external)",
        lno=0,
        # Message from ruby does not need args.
        msg=message,
        args=[],
        # The exception information makes not sense here.
        exc_info=None,
    )

    # A value of attribute relativelyCreated is in logging module calculated by
    # this way:
    # self.relativeCreated = (self.created - _startTime) * 1000
    # To update it, we need to reduce it by difference between current value
    # of attribute created (which is newer, so higher) and the correct one
    # (which commes from an external source)
    record.relativeCreated -= (record.created - created) * 1000
    record.created = created
    record.msec = usecs // 1000
    record.pcsd_group_id = str(group_id).zfill(5)
    pcsd.handle(record)

class Formatter(logging.Formatter):
    default_time_format = '%Y-%m-%dT%H:%M:%S'
    # It produces `datetime.miliseconds`
    default_msec_format = '%s.%03d'

    def __init__(self):
        super().__init__(
            fmt="{levelname[0]}, [{asctime} #{pcsd_group_id}]"
                " {levelname:>8s} -- : {message}"
            ,
            datefmt=None,
            style="{"
        )

    def format(self, record):
        # Non-external records, which are currently the minority, needs to be
        # extended by group_id (see init arugument fmt and function
        # from_external_source).
        if not hasattr(record, "pcsd_group_id"):
            record.pcsd_group_id = "00000"
        return super().format(record)

def setup(log_file):
    handler = logging.handlers.WatchedFileHandler(log_file, encoding="utf8")
    handler.setFormatter(Formatter())

    for logger_name in LOGGER_NAMES:
        pcsd_log = logging.getLogger(logger_name)
        pcsd_log.addHandler(handler)
        pcsd_log.setLevel(logging.INFO)

def enable_debug():
    # Debug messages won't be written if we call setLevel(logging.DEBUG) on the
    # handler when loggers itself have an higher level. So the level is set to
    # the loggers. Our handler has the implicit NOTSET level and there is no
    # reason to make it more complicated currently.
    for logger_name in LOGGER_NAMES:
        logging.getLogger(logger_name).setLevel(logging.DEBUG)



from unittest import mock, TestCase

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import InputModifiers
from pcs.cli.constraint_ticket import command

def _modifiers(force=True):
    options = {}
    if force:
        options["--force"] = ""
    return InputModifiers(options)

class AddTest(TestCase):
    # pylint: disable=no-self-use
    @mock.patch("pcs.cli.constraint_ticket.command.parse_args.parse_add")
    def test_call_library_with_correct_attrs(self, mock_parse_add):
        mock_parse_add.return_value = (
            "ticket", "resource_id", "", {"loss-policy": "fence"}
        )
        lib = mock.MagicMock()
        lib.constraint_ticket = mock.MagicMock()
        lib.constraint_ticket.add = mock.MagicMock()

        command.add(lib, ["argv"], _modifiers())

        mock_parse_add.assert_called_once_with(["argv"])
        lib.constraint_ticket.add.assert_called_once_with(
            "ticket", "resource_id", {"loss-policy": "fence"},
            resource_in_clone_alowed=True,
            duplication_alowed=True,
        )

    @mock.patch("pcs.cli.constraint_ticket.command.parse_args.parse_add")
    def test_refuse_resource_role_in_options(self, mock_parse_add):
        mock_parse_add.return_value = (
            "ticket", "resource_id", "resource_role", {"rsc-role": "master"}
        )
        lib = None
        self.assertRaises(
            CmdLineInputError,
            lambda: command.add(lib, ["argv"], _modifiers())
        )

    @mock.patch("pcs.cli.constraint_ticket.command.parse_args.parse_add")
    def test_put_resource_role_to_options_for_library(self, mock_parse_add):
        mock_parse_add.return_value = (
            "ticket", "resource_id", "resource_role", {"loss-policy": "fence"}
        )
        lib = mock.MagicMock()
        lib.constraint_ticket = mock.MagicMock()
        lib.constraint_ticket.add = mock.MagicMock()

        command.add(lib, ["argv"], _modifiers())

        mock_parse_add.assert_called_once_with(["argv"])
        lib.constraint_ticket.add.assert_called_once_with(
            "ticket", "resource_id",
            {"loss-policy": "fence", "rsc-role": "resource_role"},
            resource_in_clone_alowed=True,
            duplication_alowed=True,
        )

class RemoveTest(TestCase):
    def test_refuse_args_count(self):
        self.assertRaises(CmdLineInputError, lambda: command.remove(
            mock.MagicMock(),
            ["TICKET"],
            _modifiers(False),
        ))
        self.assertRaises(CmdLineInputError, lambda: command.remove(
            mock.MagicMock(),
            ["TICKET", "RESOURCE", "SOMETHING_ELSE"],
            _modifiers(False),
        ))

    def test_call_library_remove_with_correct_attrs(self):
        # pylint: disable=no-self-use
        lib = mock.MagicMock(
            constraint_ticket=mock.MagicMock(remove=mock.Mock())
        )
        command.remove(lib, ["TICKET", "RESOURCE"], _modifiers(False))
        lib.constraint_ticket.remove.assert_called_once_with(
            "TICKET", "RESOURCE",
        )



from textwrap import dedent
from unittest import TestCase

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.custom_mock import patch_getaddrinfo

from pcs.common import report_codes
from pcs.lib.commands import cluster

class UpdateLink(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.env.set_known_nodes(["node1", "node2"])
        self.link_options = {"mcastport": "12345"}
        self.node_addr_map = {"node2": "node2-addr2a"}
        self.before = dedent("""\
            totem {
                transport: knet

                interface {
                    linknumber: 2
                    mcastport: 1234
                }
            }

            nodelist {
                node {
                    ring0_addr: node1-addr0
                    ring2_addr: node1-addr2
                    name: node1
                    nodeid: 1
                }

                node {
                    ring0_addr: node2-addr0
                    ring2_addr: node2-addr2
                    name: node2
                    nodeid: 2
                }
            }
            """
        )
        self.after = dedent("""\
            totem {
                transport: knet

                interface {
                    linknumber: 2
                    mcastport: 12345
                }
            }

            nodelist {
                node {
                    ring0_addr: node1-addr0
                    ring2_addr: node1-addr2
                    name: node1
                    nodeid: 1
                }

                node {
                    ring0_addr: node2-addr0
                    ring2_addr: node2-addr2a
                    name: node2
                    nodeid: 2
                }
            }
            """
        )

    def test_not_live(self):
        (self.config
            .env.set_corosync_conf_data("")
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.update_link(
                self.env_assist.get_env(),
                "0",
                {},
                {}
            ),
            [
                fixture.error(
                    report_codes.LIVE_ENVIRONMENT_REQUIRED,
                    forbidden_options=["COROSYNC_CONF"]
                ),
            ],
            expected_in_processor=False
        )

    def test_missing_input_data(self):
        (self.config
            .corosync_conf.load_content(self.before)
            .env.push_corosync_conf(corosync_conf_text=self.before)
        )
        cluster.update_link(
            self.env_assist.get_env(),
            "0",
            {},
            {}
        )
        # Reports from pushing corosync.conf are produced in env. That code is
        # hidden in self.config.env.push_corosync_conf.
        self.env_assist.assert_reports([])

    def test_offline_nodes(self):
        patch_getaddrinfo(self, self.node_addr_map.values())
        (self.config
            .corosync_conf.load_content(self.before)
            .env.push_corosync_conf(
                corosync_conf_text=self.after,
                skip_offline_targets=True,
            )
        )
        cluster.update_link(
            self.env_assist.get_env(),
            "2",
            self.node_addr_map,
            self.link_options,
            force_flags=[report_codes.SKIP_OFFLINE_NODES]
        )

    def test_unresolvable_addresses(self):
        (self.config
            .corosync_conf.load_content(self.before)
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.update_link(
                self.env_assist.get_env(),
                "2",
                self.node_addr_map,
                self.link_options
            ),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_ADDRESSES_UNRESOLVABLE,
                    force_code=report_codes.FORCE_NODE_ADDRESSES_UNRESOLVABLE,
                    address_list=list(self.node_addr_map.values()),
                )
            ]
        )

    def test_unresolvable_addresses_forced(self):
        (self.config
            .corosync_conf.load_content(self.before)
            .env.push_corosync_conf(corosync_conf_text=self.after)
        )
        cluster.update_link(
            self.env_assist.get_env(),
            "2",
            self.node_addr_map,
            self.link_options,
            force_flags=[report_codes.FORCE]
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.NODE_ADDRESSES_UNRESOLVABLE,
                    address_list=list(self.node_addr_map.values()),
                )
            ]
        )

class UpdateLinkKnet(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.env.set_known_nodes(["node1", "node2"])
        self.before = dedent("""\
            totem {
                transport: knet

                interface {
                    linknumber: 2
                    mcastport: 1234
                    knet_transport: sctp
                }
            }

            nodelist {
                node {
                    ring0_addr: node1-addr0
                    ring2_addr: node1-addr2
                    name: node1
                    nodeid: 1
                }

                node {
                    ring0_addr: node2-addr0
                    ring2_addr: node2-addr2
                    name: node2
                    nodeid: 2
                }
            }
            """
        )

    def test_success(self):
        after = dedent("""\
            totem {
                transport: knet

                interface {
                    linknumber: 2
                    knet_transport: udp
                    knet_link_priority: 10
                }
            }

            nodelist {
                node {
                    ring0_addr: node1-addr0
                    ring2_addr: node1-addr2
                    name: node1
                    nodeid: 1
                }

                node {
                    ring0_addr: node2-addr0
                    ring2_addr: node2-addr2a
                    name: node2
                    nodeid: 2
                }
            }
            """
        )
        patch_getaddrinfo(self, ["node2-addr2a"])
        (self.config
            .corosync_conf.load_content(self.before)
            .env.push_corosync_conf(corosync_conf_text=after)
        )
        cluster.update_link(
            self.env_assist.get_env(),
            "2",
            {"node2": "node2-addr2a"},
            {"mcastport": "", "transport": "udp", "link_priority": "10"}
        )
        # Reports from pushing corosync.conf are produced in env. That code is
        # hidden in self.config.env.push_corosync_conf.
        self.env_assist.assert_reports([])

    def test_validation(self):
        patch_getaddrinfo(self, ["node2-addr0"])
        (self.config
            .corosync_conf.load_content(self.before)
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.update_link(
                self.env_assist.get_env(),
                "2",
                {
                    "nodeX": "addr-new",
                    "node2": "",
                    "node1": "node2-addr0",
                },
                {
                    "wrong": "option",
                    "transport": "unknown",
                    "link_priority": 10,
                }
            ),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="unknown",
                    option_name="transport",
                    allowed_values=("sctp", "udp"),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["wrong"],
                    option_type="link",
                    allowed=[
                        "link_priority", "mcastport", "ping_interval",
                        "ping_precision", "ping_timeout", "pong_count",
                        "transport",
                    ],
                    allowed_patterns=[],
                ),
                fixture.error(
                    report_codes.NODE_NOT_FOUND,
                    node="nodeX",
                    searched_types=[],
                ),
                fixture.error(
                    report_codes.NODE_ADDRESSES_CANNOT_BE_EMPTY,
                    node_name_list=["node2"],
                ),
                fixture.error(
                    report_codes.NODE_ADDRESSES_UNRESOLVABLE,
                    force_code=report_codes.FORCE_NODE_ADDRESSES_UNRESOLVABLE,
                    address_list=["addr-new"],
                ),
                fixture.error(
                    report_codes.NODE_ADDRESSES_ALREADY_EXIST,
                    address_list=["node2-addr0"],
                ),
            ]
        )

class UpdateLinkUdp(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.env.set_known_nodes(["node1", "node2", "node3"])
        self.before = dedent("""\
            totem {
                transport: udp

                interface {
                    broadcast: yes
                    mcastport: 1234
                    ttl: 128
                }
            }

            nodelist {
                node {
                    ring0_addr: node1-addr0
                    name: node1
                    nodeid: 1
                }

                node {
                    ring0_addr: node2-addr0
                    name: node2
                    nodeid: 2
                }

                node {
                    ring0_addr: node3-addr0
                    name: node3
                    nodeid: 3
                }
            }
            """
        )

    def test_success(self):
        after = dedent("""\
            totem {
                transport: udp

                interface {
                    ttl: 128
                    mcastaddr: 225.0.0.1
                }
            }

            nodelist {
                node {
                    ring0_addr: node1-addr0
                    name: node1
                    nodeid: 1
                }

                node {
                    ring0_addr: node2-addrA
                    name: node2
                    nodeid: 2
                }

                node {
                    ring0_addr: node3-addr0
                    name: node3
                    nodeid: 3
                }
            }
            """
        )
        patch_getaddrinfo(self, ["node2-addrA"])
        (self.config
            .corosync_conf.load_content(self.before)
            .env.push_corosync_conf(corosync_conf_text=after)
        )
        cluster.update_link(
            self.env_assist.get_env(),
            "0",
            {"node2": "node2-addrA"},
            {"mcastport": "", "broadcast": "0", "mcastaddr": "225.0.0.1"}
        )
        # Reports from pushing corosync.conf are produced in env. That code is
        # hidden in self.config.env.push_corosync_conf.
        self.env_assist.assert_reports([])

    def test_validation(self):
        patch_getaddrinfo(self, ["node3-addr0"])
        (self.config
            .corosync_conf.load_content(self.before)
        )
        self.env_assist.assert_raise_library_error(
            lambda: cluster.update_link(
                self.env_assist.get_env(),
                "0",
                {
                    "nodeX": "addr-new",
                    "node2": "",
                    "node1": "node3-addr0",
                },
                {
                    "wrong": "option",
                    "broadcast": "1",
                    "mcastaddr": "address",
                }
            ),
            []
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="address",
                    option_name="mcastaddr",
                    allowed_values="an IP address",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["wrong"],
                    option_type="link",
                    allowed=[
                        "bindnetaddr", "broadcast", "mcastaddr", "mcastport",
                        "ttl",
                    ],
                    allowed_patterns=[],
                ),
                fixture.error(
                    report_codes.PREREQUISITE_OPTION_MUST_BE_DISABLED,
                    option_name="mcastaddr",
                    option_type="link",
                    prerequisite_name="broadcast",
                    prerequisite_type="link",
                ),
                fixture.error(
                    report_codes.NODE_NOT_FOUND,
                    node="nodeX",
                    searched_types=[],
                ),
                fixture.error(
                    report_codes.NODE_ADDRESSES_CANNOT_BE_EMPTY,
                    node_name_list=["node2"],
                ),
                fixture.error(
                    report_codes.NODE_ADDRESSES_UNRESOLVABLE,
                    force_code=report_codes.FORCE_NODE_ADDRESSES_UNRESOLVABLE,
                    address_list=["addr-new"],
                ),
                fixture.error(
                    report_codes.NODE_ADDRESSES_ALREADY_EXIST,
                    address_list=["node3-addr0"],
                ),
            ]
        )



from pcs.lib import resource_agent


def list_standards(lib_env):
    """
    List resource agents standards (ocf, lsb, ... ) on the local host
    """
    return resource_agent.list_resource_agents_standards(lib_env.cmd_runner())


def list_ocf_providers(lib_env):
    """
    List resource agents ocf providers on the local host
    """
    return resource_agent.list_resource_agents_ocf_providers(
        lib_env.cmd_runner()
    )


def list_agents_for_standard_and_provider(lib_env, standard_provider=None):
    """
    List resource agents for specified standard on the local host
    string standard_provider standard[:provider], e.g. None, ocf, ocf:pacemaker
    """
    if standard_provider:
        standards = [standard_provider]
    else:
        standards = resource_agent.list_resource_agents_standards(
            lib_env.cmd_runner()
        )
    agents = []
    for std in standards:
        agents += resource_agent.list_resource_agents(lib_env.cmd_runner(), std)
    return sorted(
        agents,
        # works with both str and unicode in both python 2 and 3
        key=lambda x: x.lower()
    )


def list_agents(lib_env, describe=True, search=None):
    """
    List all resource agents on the local host, optionally filtered and
        described
    bool describe load and return agents' description as well
    string search return only agents which name contains this string
    """
    runner = lib_env.cmd_runner()

    # list agents for all standards and providers
    agent_names = []
    for std in resource_agent.list_resource_agents_standards_and_providers(
        runner
    ):
        agent_names += [
            "{0}:{1}".format(std, agent)
            for agent in resource_agent.list_resource_agents(runner, std)
        ]
    agent_names.sort(
        # works with both str and unicode in both python 2 and 3
        key=lambda x: x.lower()
    )
    return _complete_agent_list(
        runner,
        agent_names,
        describe,
        search,
        resource_agent.ResourceAgent
    )


def _complete_agent_list(
    runner, agent_names, describe, search, metadata_class
):
    # filter agents by name if requested
    if search:
        search_lower = search.lower()
        agent_names = [
            name for name in agent_names if search_lower in name.lower()
        ]

    # complete the output and load descriptions if requested
    agent_list = []
    for name in agent_names:
        try:
            agent_metadata = metadata_class(runner, name)
            if describe:
                agent_list.append(agent_metadata.get_description_info())
            else:
                agent_list.append(agent_metadata.get_name_info())
        except resource_agent.ResourceAgentError:
            #we don't return it in the list:
            #
            #UnableToGetAgentMetadata - if we cannot get valid metadata, it's
            #not a resource agent
            #
            #InvalidResourceAgentName - invalid name cannot be used with a new
            #resource. The list of names is gained from "crm_resource" whilst
            #pcs is doing the validation. So there can be a name that pcs does
            #not recognize as valid.
            #
            #Providing a warning is not the way (currently). Other components
            #read this list and do not expect warnings there. Using the stderr
            #(to separate warnings) is currently difficult.
            pass
    return agent_list


def describe_agent(lib_env, agent_name):
    """
    Get agent's description (metadata) in a structure
    string agent_name name of the agent
    """
    agent = resource_agent.find_valid_resource_agent_by_name(
        lib_env.report_processor,
        lib_env.cmd_runner(),
        agent_name,
        absent_agent_supported=False
    )
    return agent.get_full_info()



import inspect
import logging
import os
import os.path
import sys
from functools import partial

from unittest import mock

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_raise_library_error, prepare_diff
from pcs_test.tools.case_analysis import test_failed
from pcs_test.tools.command_env import spy
from pcs_test.tools.command_env.calls import Queue as CallQueue
from pcs_test.tools.command_env.config import Config
from pcs_test.tools.command_env.mock_fs import(
    get_fs_mock,
    is_fs_call_in,
)
from pcs_test.tools.command_env.mock_get_local_corosync_conf import(
    get_get_local_corosync_conf
)
from pcs_test.tools.command_env.mock_node_communicator import NodeCommunicator
from pcs_test.tools.command_env.mock_push_cib import(
    get_push_cib,
    is_push_cib_call_in,
)
from pcs_test.tools.command_env.mock_push_corosync_conf import(
    get_push_corosync_conf,
    is_push_corosync_conf_call_in,
)
from pcs_test.tools.command_env.mock_raw_file import get_raw_file_mock
from pcs_test.tools.command_env.mock_runner import Runner
from pcs_test.tools.custom_mock import MockLibraryReportProcessor

from pcs.common.file import RawFile
from pcs.common.node_communicator import NodeCommunicatorFactory
from pcs.lib.env import LibraryEnvironment

patch_lib_env = partial(mock.patch.object, LibraryEnvironment)

def patch_env(call_queue, config, init_env, patch_is_systemd=True):
    #It is mandatory to patch some env objects/methods. It is ok when command
    #does not use this objects/methods and specify no call for it. But it would
    #be a problem when the test succeded because the live call respond correctly
    #by accident. Such test would fails on different machine (with another live
    #environment)

    get_cmd_runner = init_env.cmd_runner
    get_node_communicator = init_env.get_node_communicator
    mock_communicator_factory = mock.Mock(spec_set=NodeCommunicatorFactory)
    mock_communicator_factory.get_communicator = (
        # TODO: use request_timeout
        lambda request_timeout=None:
            NodeCommunicator(call_queue) if not config.spy
            else spy.NodeCommunicator(get_node_communicator())
    )
    patcher_list = [
        patch_lib_env(
            "cmd_runner",
            lambda env:
            spy.Runner(get_cmd_runner()) if config.spy else Runner(
                call_queue,
                env_vars={} if not config.env.cib_tempfile else {
                    "CIB_file": config.env.cib_tempfile,
                }
            )
        ),

        mock.patch(
            "pcs.lib.env.get_local_corosync_conf",
            get_get_local_corosync_conf(call_queue) if not config.spy
                else spy.get_local_corosync_conf
        ),

        patch_lib_env("communicator_factory", mock_communicator_factory),
    ]
    if patch_is_systemd:
        # In all the tests we assume that we are running on top of a systemd
        # running system. If needed, this may be turned off for some particular
        # tests. Note that the patched function is cached therefore is patched
        # here and not in every tests.
        patcher_list.append(
            mock.patch("pcs.lib.external.is_systemctl", lambda: True)
        )

    if is_fs_call_in(call_queue):
        fs_mock = get_fs_mock(call_queue)
        builtin = (
            ("__builtin__" if sys.version_info[0] == 2 else "builtins")+".{0}"
        ).format

        patcher_list.extend([
            mock.patch(
                builtin("open"),
                fs_mock("open", open)
            ),
            mock.patch(
                "os.path.exists",
                fs_mock("os.path.exists", os.path.exists)
            ),
            mock.patch(
                "os.path.isdir",
                fs_mock("os.path.isdir", os.path.isdir)
            ),
            mock.patch(
                "os.path.isfile",
                fs_mock("os.path.isfile", os.path.isfile)
            ),
            mock.patch(
                "os.listdir",
                fs_mock("os.listdir", os.listdir)
            ),
            mock.patch(
                "os.chmod",
                fs_mock("os.chmod", os.chmod)
            ),
            mock.patch(
                "os.chown",
                fs_mock("os.chown", os.chown)
            ),
        ])

    raw_file_mock = get_raw_file_mock(call_queue)
    for method_name, dummy_method in inspect.getmembers(
        RawFile, inspect.isfunction
    ):
        # patch all public methods
        # inspect.isfunction must be used instead of ismethod because we are
        # working with a class and not an instance - no method is bound yet so
        # it would return an empty list
        # "protected" methods start with _
        # "private" methods start with _<class_name>__
        if method_name.startswith("_"):
            continue
        patcher_list.append(
            mock.patch.object(
                RawFile,
                method_name,
                getattr(raw_file_mock, method_name)
            )
        )

    # It is not always desirable to patch these methods. Some tests may patch
    # only the internals (runner etc.). So these methods are only patched when
    # it is explicitly configured.
    if is_push_cib_call_in(call_queue):
        patcher_list.append(
            patch_lib_env("push_cib", get_push_cib(call_queue))
        )
    if is_push_corosync_conf_call_in(call_queue):
        patcher_list.append(
            patch_lib_env(
                "push_corosync_conf",
                get_push_corosync_conf(call_queue)
            )
        )

    for patcher in patcher_list:
        patcher.start()

    def unpatch():
        for patcher in patcher_list:
            patcher.stop()

    return unpatch

class EnvAssistant:
    # pylint: disable=too-many-instance-attributes
    def __init__(
        self, config=None, test_case=None,
        exception_reports_in_processor_by_default=True
    ):
        """
        TestCase test_case -- cleanup callback is registered to test_case if is
            provided
        """
        self.__call_queue = None
        self.__config = config if config else Config()
        self.__reports_asserted = False
        self.__extra_reports = []
        self.exception_reports_in_processor_by_default = (
            exception_reports_in_processor_by_default
        )

        self.__unpatch = None
        self.__original_mocked_corosync_conf = None

        if test_case:
            test_case.addCleanup(lambda: self.cleanup(test_case))

    @property
    def config(self):
        return self.__config

    def cleanup(self, current_test):
        if self.__unpatch:
            self.__unpatch()

        if test_failed(current_test):
            # We have already got the message that main test failed. There is
            # a high probability that something remains in reports or in the
            # queue etc. But it is only consequence of the main test fail. And
            # we do not want to make the report confusing.
            return

        if not self.__reports_asserted:
            self.__assert_environment_created()
            if not self.__config.spy:
                self._env.report_processor.assert_reports(
                    self.__extra_reports,
                    hint="EnvAssistant.cleanup - is param"
                        " 'expected_in_processor' in the method"
                        " 'assert_raise_library_error' set correctly?"
                )

        if not self.__config.spy:
            if self.__call_queue and self.__call_queue.remaining:
                raise AssertionError(
                    "There are remaining expected calls: \n    '{0}'"
                    .format("'\n    '".join([
                        repr(call) for call in self.__call_queue.remaining
                    ]))
                )
            # If pushing corosync.conf has not been patched in the
            # LibraryEnvironment and the LibraryEnvironment was constructed
            # with a mocked corosync.conf, check if it was changed without the
            # change being specified in a test.
            # If no env.push_corosync_conf call has been specified, no mocking
            # occurs, any changes to corosync.conf are done just in memory and
            # nothing gets reported. So an explicit check is necessary.
            corosync_conf_orig = self.__original_mocked_corosync_conf
            # pylint: disable=protected-access
            corosync_conf_env = self._env._corosync_conf_data
            if (
                corosync_conf_orig
                and
                corosync_conf_orig != corosync_conf_env
            ):
                raise AssertionError(
                    (
                        "An unexpected change to corosync.conf in "
                        "LibraryEnvironment has been detected:\n{0}"
                    ).format(
                        prepare_diff(corosync_conf_orig, corosync_conf_env)
                    )
                )


    def get_env(self, patch_is_systemd=True):
        self.__call_queue = CallQueue(self.__config.calls)
        #pylint: disable=attribute-defined-outside-init
        self._env = LibraryEnvironment(
            mock.MagicMock(logging.Logger),
            MockLibraryReportProcessor(),
            cib_data=self.__config.env.cib_data,
            corosync_conf_data=self.__config.env.corosync_conf_data,
            known_hosts_getter=(
                (lambda: self.__config.spy.known_hosts) if self.__config.spy
                else self.__config.env.known_hosts_getter
            ),
            booth_files_data=self.__config.env.booth,
        )
        self.__unpatch = patch_env(
            self.__call_queue, self.__config, self._env,
            patch_is_systemd=patch_is_systemd
        )
        # If pushing corosync.conf has not been patched in the
        # LibraryEnvironment, store any corosync.conf passed to the
        # LibraryEnvironment for check for changes in cleanup.
        if not is_push_corosync_conf_call_in(self.__call_queue):
            self.__original_mocked_corosync_conf = (
                self.__config.env.corosync_conf_data
            )
        return self._env

    def assert_reports(self, expected_reports):
        self.__reports_asserted = True
        self.__assert_environment_created()
        self._env.report_processor.assert_reports(
            (
                expected_reports.reports
                    if isinstance(expected_reports, fixture.ReportStore)
                else expected_reports
            )
            +
            self.__extra_reports
        )

    def assert_raise_library_error(
        self, command, reports=None, expected_in_processor=None
    ):
        if reports is None:
            reports = []

        if not isinstance(reports, list):
            raise self.__list_of_reports_expected(reports)

        if expected_in_processor is None:
            expected_in_processor = (
                self.exception_reports_in_processor_by_default
            )

        assert_raise_library_error(command, *reports)
        if expected_in_processor:
            self.__extra_reports = reports

    def __assert_environment_created(self):
        if not hasattr(self, "_env"):
            raise AssertionError(
                "LibraryEnvironment was not created in EnvAssitant."
                " Have you called method get_env?"
            )

    def __list_of_reports_expected(self, reports):
        return AssertionError(
            "{0}.{1} expects 'list' as reports parameter, '{2}' was given"
            .format(
                self.__class__.__name__,
                "assert_raise",
                type(reports).__name__
            )
        )



from functools import partial

from pcs.lib import reports
from pcs.lib.cib.constraint import constraint
from pcs.lib.cib.tools import check_new_id_applicable
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.values import is_score, SCORE_INFINITY

TAG_NAME = 'rsc_colocation'
DESCRIPTION = "constraint id"
SCORE_NAMES = ("score", "score-attribute", "score-attribute-mangle")

def prepare_options_with_set(cib, options, resource_set_list):
    options = constraint.prepare_options(
        tuple(SCORE_NAMES),
        options,
        partial(constraint.create_id, cib, TAG_NAME, resource_set_list),
        partial(check_new_id_applicable, cib, DESCRIPTION),
    )

    if "score" in options and not is_score(options["score"]):
        raise LibraryError(reports.invalid_score(options["score"]))

    score_attrs_count = len([
        name for name in options.keys() if name in SCORE_NAMES
    ])
    if score_attrs_count > 1:
        raise LibraryError(reports.multiple_score_options())

    if score_attrs_count == 0:
        options["score"] = SCORE_INFINITY

    return options



import abc

from pcs.lib.errors import ReportItem, ReportItemList


class SimpleReportProcessorInterface(abc.ABC):
    def report(self, report_item: ReportItem) -> ReportItemList:
        return self.report_list([report_item])

    @abc.abstractmethod
    def report_list(self, report_list: ReportItemList) -> ReportItemList:
        raise NotImplementedError()


class SimpleReportProcessor(SimpleReportProcessorInterface):
    """
    This class is a wrapper for a report processor class and at the same time
    implements interface of a simple report processor. This class interface for
    easy checking if some errors have been reported.
    """
    def __new__(cls, report_processor: SimpleReportProcessorInterface):
        if isinstance(report_processor, cls):
            # There is no point in wrapping the same object multiple times
            return report_processor
        self = super().__new__(cls)
        self.__init__(report_processor)
        return self

    def __init__(self, report_processor: SimpleReportProcessorInterface):
        self._report_processor = report_processor
        self._error_list: ReportItemList = []

    def report_list(self, report_list: ReportItemList) -> ReportItemList:
        error_list = self._report_processor.report_list(report_list)
        self._error_list.extend(error_list)
        return error_list

    @property
    def has_errors(self) -> bool:
        return bool(self._error_list)

    def clear_errors(self) -> None:
        self._error_list = []



import os
from unittest import TestCase

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_raise_library_error

from pcs import settings
from pcs.common import (
    file_type_codes,
    report_codes,
)
from pcs.lib.booth import env
from pcs.lib.file.raw_file import GhostFile


class BoothEnv(TestCase):
    def test_ghost_conf_real_key(self):
        # pylint: disable=no-self-use
        assert_raise_library_error(
            lambda: env.BoothEnv(
                "my_booth",
                {"config_data": "some config data".encode("utf-8")}
            ),
            fixture.error(
                report_codes.LIVE_ENVIRONMENT_NOT_CONSISTENT,
                mocked_files=[file_type_codes.BOOTH_CONFIG],
                required_files=[file_type_codes.BOOTH_KEY],
            ),
        )

    def test_real_conf_ghost_key(self):
        # pylint: disable=no-self-use
        assert_raise_library_error(
            lambda: env.BoothEnv(
                "my_booth",
                {"key_data": "some key data".encode("utf-8")}
            ),
            fixture.error(
                report_codes.LIVE_ENVIRONMENT_NOT_CONSISTENT,
                mocked_files=[file_type_codes.BOOTH_KEY],
                required_files=[file_type_codes.BOOTH_CONFIG],
            ),
        )

    def test_real(self):
        my_env = env.BoothEnv("my_booth", {})
        self.assertEqual("my_booth", my_env.instance_name)
        self.assertFalse(isinstance(my_env.config.raw_file, GhostFile))
        self.assertFalse(isinstance(my_env.key.raw_file, GhostFile))
        self.assertEqual(
            os.path.join(settings.booth_config_dir, "my_booth.conf"),
            my_env.config_path
        )
        self.assertEqual(
            os.path.join(settings.booth_config_dir, "my_booth.key"),
            my_env.key_path
        )
        self.assertEqual([], my_env.ghost_file_codes)
        self.assertEqual({}, my_env.export())

        site_list = ["site1", "site2"]
        arbitrator_list = ["arbitrator1"]
        facade = my_env.create_facade(site_list, arbitrator_list)
        self.assertEqual(site_list, facade.get_sites())
        self.assertEqual(arbitrator_list, facade.get_arbitrators())


    def test_ghost(self):
        config_data = "some config_data".encode("utf-8")
        key_data = "some key_data".encode("utf-8")
        key_path = "some key path"
        my_env = env.BoothEnv(
            "my_booth",
            {
                "config_data": config_data,
                "key_data": key_data,
                "key_path": key_path,
            }
        )
        self.assertEqual("my_booth", my_env.instance_name)
        self.assertTrue(isinstance(my_env.config.raw_file, GhostFile))
        self.assertTrue(isinstance(my_env.key.raw_file, GhostFile))
        with self.assertRaises(AssertionError) as cm:
            dummy_path = my_env.config_path
        self.assertEqual(
            "Reading config path is supported only in live environment",
            str(cm.exception)
        )
        self.assertEqual(key_path, my_env.key_path)
        self.assertEqual(
            [file_type_codes.BOOTH_CONFIG, file_type_codes.BOOTH_KEY],
            my_env.ghost_file_codes
        )
        self.assertEqual(
            {
                "config_file": {"content": config_data},
                "key_file": {"content": key_data},
            },
            my_env.export()
        )

        site_list = ["site1", "site2"]
        arbitrator_list = ["arbitrator1"]
        facade = my_env.create_facade(site_list, arbitrator_list)
        self.assertEqual(site_list, facade.get_sites())
        self.assertEqual(arbitrator_list, facade.get_arbitrators())

    def test_invalid_instance(self):
        # pylint: disable=no-self-use
        assert_raise_library_error(
            lambda: env.BoothEnv("/tmp/booth/booth", {}),
            fixture.error(
                report_codes.BOOTH_INVALID_NAME,
                name="/tmp/booth/booth",
                reason="contains illegal character '/'",
            ),
        )

    def test_invalid_instance_ghost(self):
        # pylint: disable=no-self-use
        assert_raise_library_error(
            lambda: env.BoothEnv(
                "../../booth/booth",
                {
                    "config_data": "some config data",
                    "key_data": "some key data",
                    "key_path": "some key path",
                }
            ),
            fixture.error(
                report_codes.BOOTH_INVALID_NAME,
                name="../../booth/booth",
                reason="contains illegal character '/'",
            ),
        )

    def test_default_instance(self):
        self.assertEqual(
            env.BoothEnv(None, {}).instance_name,
            "booth"
        )



import os.path

systemctl_binary = "/bin/systemctl"
chkconfig_binary = "/sbin/chkconfig"
service_binary = "/sbin/service"
pacemaker_binaries = "/usr/sbin/"
crm_resource_binary = os.path.join(pacemaker_binaries, "crm_resource")
corosync_binaries = "/usr/sbin/"
corosync_qnet_binaries = "/usr/bin/"
corosync_conf_dir = "/etc/corosync/"
corosync_conf_file = os.path.join(corosync_conf_dir, "corosync.conf")
corosync_uidgid_dir = os.path.join(corosync_conf_dir, "uidgid.d/")
corosync_qdevice_net_server_certs_dir = os.path.join(
    corosync_conf_dir,
    "qnetd/nssdb"
)
corosync_qdevice_net_client_certs_dir = os.path.join(
    corosync_conf_dir,
    "qdevice/net/nssdb"
)
corosync_qdevice_net_client_ca_file_name = "qnetd-cacert.crt"
corosync_authkey_file = os.path.join(corosync_conf_dir, "authkey")
# Must be set to 256 for corosync to work in FIPS environment.
corosync_authkey_bytes = 256
corosync_log_file = "/var/log/cluster/corosync.log"
pacemaker_authkey_file = "/etc/pacemaker/authkey"
# Using the same value as for corosync. Higher values MAY work in FIPS.
pacemaker_authkey_bytes = 256
booth_authkey_file_mode = 0o600
# Booth does not support keys longer than 64 bytes.
booth_authkey_bytes = 64
cluster_conf_file = "/etc/cluster/cluster.conf"
fence_agent_binaries = "/usr/sbin/"
pacemaker_schedulerd = "/usr/libexec/pacemaker/pacemaker-schedulerd"
pacemaker_controld = "/usr/libexec/pacemaker/pacemaker-controld"
pacemaker_based = "/usr/libexec/pacemaker/pacemaker-based"
pacemaker_fenced = "/usr/libexec/pacemaker/pacemaker-fenced"
pcs_version = "0.10.4"
crm_report = os.path.join(pacemaker_binaries, "crm_report")
crm_rule = os.path.join(pacemaker_binaries, "crm_rule")
crm_verify = os.path.join(pacemaker_binaries, "crm_verify")
cibadmin = os.path.join(pacemaker_binaries, "cibadmin")
crm_mon_schema = '/usr/share/pacemaker/crm_mon.rng'
agent_metadata_schema = "/usr/share/resource-agents/ra-api-1.dtd"
pcsd_var_location = "/var/lib/pcsd/"
pcsd_cert_location = os.path.join(pcsd_var_location, "pcsd.crt")
pcsd_key_location = os.path.join(pcsd_var_location, "pcsd.key")
pcsd_known_hosts_location = os.path.join(pcsd_var_location, "known-hosts")
pcsd_users_conf_location = os.path.join(pcsd_var_location, "pcs_users.conf")
pcsd_settings_conf_location = os.path.join(
    pcsd_var_location, "pcs_settings.conf"
)
pcsd_exec_location = "/usr/lib/pcsd/"
pcsd_log_location = "/var/log/pcsd/pcsd.log"
pcsd_default_port = 2224
pcsd_config = "/etc/sysconfig/pcsd"
cib_dir = "/var/lib/pacemaker/cib/"
pacemaker_uname = "hacluster"
pacemaker_gname = "haclient"
sbd_binary = "/usr/sbin/sbd"
sbd_watchdog_default = "/dev/watchdog"
sbd_config = "/etc/sysconfig/sbd"
# this limit is also mentioned in docs, change there as well
sbd_max_device_num = 3
# message types are also mentioned in docs, change there as well
sbd_message_types = ["test", "reset", "off", "crashdump", "exit", "clear"]
pacemaker_wait_timeout_status = 124
booth_config_dir = "/etc/booth"
booth_binary = "/usr/sbin/booth"
default_request_timeout = 60
pcs_bundled_dir = "/usr/lib/pcs/bundled/"
pcs_bundled_pacakges_dir = os.path.join(pcs_bundled_dir, "packages")

default_ssl_ciphers = "DEFAULT:!RC4:!3DES:@STRENGTH"

# Ssl options are based on default options in python (maybe with some extra
# options). Format here is the same as the PCSD_SSL_OPTIONS environment
# variable format (string with coma as a delimiter).
default_ssl_options = ",".join([
    "OP_NO_COMPRESSION",
    "OP_CIPHER_SERVER_PREFERENCE",
    "OP_SINGLE_DH_USE",
    "OP_SINGLE_ECDH_USE",
    "OP_NO_SSLv2",
    "OP_NO_SSLv3",
    "OP_NO_TLSv1",
    "OP_NO_TLSv1_1",
    "OP_NO_RENEGOTIATION",
])
# Set pcsd_gem_path to None if there are no bundled ruby gems and the path does
# not exists.
pcsd_gem_path = "vendor/bundle/ruby"
ruby_executable = "/usr/bin/ruby"

gui_session_lifetime_seconds = 60 * 60



def format_call(call):
    if hasattr(call, "format"):
        return call.format()
    return call

def show_calls(name_list, call_list):
    return "\n".join([
        "  {0}. '{1}': {2}".format(i, x[0], format_call(x[1]))
        for i, x in enumerate(zip(name_list, call_list))
    ])


class Queue:
    def __init__(self, call_list_builder=None):
        if not call_list_builder:
            call_list_builder = CallListBuilder()

        self.__call_list = call_list_builder.calls
        self.__name_list = call_list_builder.names

        self.__index = 0

    def take(self, type_of_call, real_call_info=None):
        if self.__index >= len(self.__call_list):
            raise self.__extra_call(type_of_call, real_call_info)

        call = self.__call_list[self.__index]

        if call.type != type_of_call:
            raise self.__unexpected_type(call, type_of_call, real_call_info)

        self.__index += 1
        return self.__index, call

    def has_type(self, call_type):
        return any(call.type == call_type for call in self.__call_list)

    @property
    def remaining(self):
        return self.__call_list[self.__index:]

    @property
    def taken(self):
        return self.__call_list[:self.__index]

    def error_with_context(self, message):
        return AssertionError(
            "{0}\nAll calls in queue (current index={1}):\n{2}".format(
                message,
                self.__index,
                show_calls(self.__name_list, self.__call_list),
            )
        )

    def __unexpected_type(self, call, real_type, real_call_info):
        return self.error_with_context(
            (
                "{0}. call was expected as '{1}' type but was '{2}' type"
                "\n  expected call: {3}{4}"
                "\nHint: check call compatibility: for example if you use"
                " env.push_cib() then runner.cib.push() will be never launched"
            ).format(
                self.__index + 1,
                call.type,
                real_type,
                call,
                "\n  real call: {0}".format(real_call_info) if real_call_info
                    else ""
                ,
            )
        )

    def __extra_call(self, type_of_call, real_call_info):
        return self.error_with_context(
            "No next call expected, but was ({0}):\n    '{1}'"
            .format(type_of_call, real_call_info)
        )

class CallListBuilder:
    def __init__(self):
        self.__call_list = []
        self.__name_list = []

    @property
    def calls(self):
        return list(self.__call_list)

    @property
    def names(self):
        return list(self.__name_list)


    def __set(self, instead_name, name, call):
        """
        Replace call that has key instead_name with new call that has key name

        string name -- key of the call
        Call call
        string instead_name -- key of call instead of which this new call is to
            be placed
        """
        if instead_name not in self.__name_list:
            raise self.__cannot_put("instead of", instead_name, name, call)

        for i, current_name in enumerate(self.__name_list):
            if current_name == instead_name:
                self.__call_list[i] = call
                #yes we change the name as well
                self.__name_list[i] = name
                return

    def __append(self, name, call):
        """
        Append call.

        string name -- key of the call
        Call call
        """
        self.__name_list.append(name)
        self.__call_list.append(call)

    def __insert(self, before_name, name, call):
        """
        Insert call before call with before_name.

        string before_name -- key of call before which this new call is to be
            placed
        string name -- key of the call
        Call call
        """
        if before_name not in self.__name_list:
            raise self.__cannot_put("before", before_name, name, call)

        index = self.__name_list.index(before_name)
        self.__name_list.insert(index, name)
        self.__call_list.insert(index, call)

    def remove(self, name):
        """
        Remove call under key name.
        """
        try:
            index = self.__name_list.index(name)
            del self.__call_list[index]
            del self.__name_list[index]
        except ValueError:
            raise self.__name_not_exists(name)

    def get(self, name):
        """
        Get first call with name.

        string name -- key of the call
        """
        try:
            return self.__call_list[self.__name_list.index(name)]
        except ValueError:
            raise self.__name_not_exists(name)

    def place(self, name, call, before=None, instead=None):
        """
        Place call into calllist.

        string name -- key of the call
        Call call
        string before -- key of call before which this new call is to be placed
        string instead -- key of call instead of which this new call is to be
            placed
        """
        if name and name in self.__name_list and instead != name:
            raise self.__name_exists_already(name)

        if before and instead:
            raise self.__cannot_use_before_and_instead(
                name,
                call,
                before,
                instead,
            )

        if not hasattr(call, "type") or not call.type:
            raise self.__type_of_call_is_not_specified(call)

        if before:
            self.__insert(before, name, call)
        elif instead:
            self.__set(instead, name, call)
        else:
            self.__append(name, call)

    def __error_with_context(self, message):
        return AssertionError(
            "{0}\nIn the confituration call collection are calls:\n{1}".format(
                message,
                show_calls(self.__name_list, self.__call_list),
            )
        )

    @staticmethod
    def __type_of_call_is_not_specified(call):
        return AssertionError(
            "Class {0}.{1} must have the attribute 'type' with no-falsy value."
            .format(call.__module__, call.__class__.__name__)
        )

    def __name_not_exists(self, name):
        return self.__error_with_context(
            "Call named '{0}' does not exist.".format(name)
        )

    def __name_exists_already(self, name):
        return self.__error_with_context(
            "Name '{0}' is in this configuration already.".format(name)
        )

    def __cannot_use_before_and_instead(self, name, call, before, instead):
        return self.__error_with_context(
            (
                "Args 'before' ({0}) and 'instead' ({1}) cannot be used"
                " together\n  '{2}': {3}"
            ).format(before, instead, name, call)
        )

    def __cannot_put(self, where_type, where_name, name, call):
        return self.__error_with_context(
            (
                "Cannot put call named '{0}' ({1}) {2} '{3}'"
                " because '{3}' does not exist."
            ).format(
                name,
                call,
                where_type,
                where_name,
            )
        )



from collections import defaultdict

from pcs.lib import reports
from pcs.lib.cib.resource import (
    clone,
    common,
    group,
    primitive,
)
from pcs.lib.cib.tools import ElementSearcher

class ValidateMoveResourcesToGroup():
    # Base class for validating moving resources to a group. It holds common
    # code and hides flags which should not be exposed to users of validating
    # classes.
    def __init__(self):
        self._group_element = None
        self._resource_element_list = []
        self._adjacent_resource_element = None

    def _validate_elements(
        self, bad_or_missing_group_specified=False,
        bad_resources_specified=False, bad_adjacent_specified=False
    ):
        return (
            self._validate_resource_elements(
                bad_or_missing_group_specified,
                bad_resources_specified,
                bad_adjacent_specified,
            )
            +
            self._validate_adjacent_resource_element()
        )

    def _validate_resource_elements(
        self, bad_or_missing_group_specified, bad_resources_specified,
        bad_adjacent_specified
    ):
        report_list = []

        # Check if the group element is really a group unless it has been
        # checked before.
        if (
            not bad_or_missing_group_specified
            and
            not group.is_group(self._group_element)
        ):
            bad_or_missing_group_specified = True
            report_list.append(
                reports.id_belongs_to_unexpected_type(
                    self._group_element.attrib.get("id"),
                    expected_types=[group.TAG],
                    current_type=self._group_element.tag
                )
            )

        # Report an error if no resources were specified. If resource ids have
        # been specified but they are not valid resource ids, then some
        # resources were specified, even though not valid ones.
        if not bad_resources_specified and not self._resource_element_list:
            report_list.append(reports.cannot_group_resource_no_resources())

        resources_already_in_the_group = set()
        resources_count = defaultdict(int)
        for resource in self._resource_element_list:
            resource_id = resource.attrib.get("id")
            if not primitive.is_primitive(resource):
                report_list.append(
                    reports.cannot_group_resource_wrong_type(
                        resource_id,
                        resource.tag
                    )
                )
                continue
            resources_count[resource_id] = resources_count[resource_id] + 1
            parent = resource.getparent()
            if parent is not None:
                if group.is_group(parent):
                    # If no adjacent resource has been specified (either valid
                    # or invalid), the resources must not already be in the
                    # group, if the group already exists.
                    if (
                        not bad_or_missing_group_specified
                        and
                        not bad_adjacent_specified
                        and
                        self._adjacent_resource_element is None
                        and
                        (
                            parent.attrib.get("id")
                            ==
                            self._group_element.attrib.get("id")
                        )
                    ):
                        resources_already_in_the_group.add(resource_id)
                elif parent.tag != "resources":
                    # If the primitive is not in a 'group' or 'resources' tag,
                    # it is either in a clone, master, bundle or similar tag
                    # and cannot be put into a group.
                    report_list.append(
                        reports.cannot_group_resource_wrong_type(
                            resource_id,
                            parent.tag
                        )
                    )
        if resources_already_in_the_group:
            report_list.append(
                reports.cannot_group_resource_already_in_the_group(
                    resources_already_in_the_group,
                    self._group_element.attrib.get("id")
                )
            )
        more_than_once_resources = [
            resource for resource, count in resources_count.items()
            if count > 1
        ]
        if more_than_once_resources:
            report_list.append(
                reports.cannot_group_resource_more_than_once(
                    more_than_once_resources
                )
            )
        return report_list

    def _validate_adjacent_resource_element(self):
        report_list = []
        if self._adjacent_resource_element is not None:
            if self._group_element is not None:
                adjacent_parent = self._adjacent_resource_element.getparent()
                if not (
                    adjacent_parent is not None
                    and
                    group.is_group(adjacent_parent)
                    and
                    (
                        adjacent_parent.attrib.get("id")
                        ==
                        self._group_element.attrib.get("id")
                    )
                ):
                    report_list.append(
                        reports
                        .cannot_group_resource_adjacent_resource_not_in_group(
                            self._adjacent_resource_element.attrib.get("id"),
                            self._group_element.attrib.get("id")
                        )
                    )
            for resource in self._resource_element_list:
                if (
                    self._adjacent_resource_element.attrib.get("id")
                    ==
                    resource.attrib.get("id")
                ):
                    report_list.append(
                        reports.cannot_group_resource_next_to_itself(
                            self._adjacent_resource_element.attrib.get("id")
                        )
                    )
                    break
        return report_list

class ValidateMoveResourcesToGroupByElements(ValidateMoveResourcesToGroup):
    """
    Validate putting resources into a group or moving them within their group
    """
    def __init__(
        self, group_element, resource_element_list,
        adjacent_resource_element=None
    ):
        """
        etree.Element group_element -- the group to put resources into
        iterable resource_element_list -- resources to put into the group
        etree.Element adjacent_resource_element -- put resources beside this one
        """
        super().__init__()
        self._group_element = group_element
        self._resource_element_list = resource_element_list
        self._adjacent_resource_element = adjacent_resource_element

    def validate(self):
        """
        Run the validation and return a report item list
        """
        return self._validate_elements()

class ValidateMoveResourcesToGroupByIds(ValidateMoveResourcesToGroup):
    """
    Validate putting resources into a group or moving them within their group
    """
    def __init__(
        self, group_id, resource_id_list, adjacent_resource_id=None
    ):
        """
        string group_id -- id of the group to put resources into
        iterable resource_id_list -- ids of resources to put into the group
        string adjacent_resource_id -- put resources beside this one
        """
        super().__init__()
        self._group_id = group_id
        self._resource_id_list = resource_id_list
        self._adjacent_resource_id = adjacent_resource_id

    def validate(self, resources_section, id_provider):
        """
        Run the validation and return a report item list

        etree.Element resources_section -- resources section of a cib
        IdProvider id_provider -- elements' ids generator and uniqueness checker
        """
        report_list = []

        # Check that group_id either matches an existing group element or is
        # not occupied by any other element.
        group_missing_id_valid = False
        group_searcher = ElementSearcher(
            group.TAG, self._group_id, resources_section
        )
        if group_searcher.element_found():
            self._group_element = group_searcher.get_element()
        elif group_searcher.validate_book_id(
            id_provider, id_description="group name"
        ):
            group_missing_id_valid = True
        else:
            report_list.extend(group_searcher.get_errors())

        # Get resource elements to move to the group.
        # Get all types of resources, so that validation can later tell for
        # example: 'C' is a clone, clones cannot be put into a group. If we
        # only searched for primitives here, we would get 'C' is not a
        # resource, which is not that informative.
        self._resource_element_list = common.find_resources_and_report(
            resources_section, self._resource_id_list, report_list
        )

        # Get an adjacent resource element.
        if self._adjacent_resource_id is not None:
            # If the group already exists, check the adjacent resource is in it.
            if self._group_element is not None:
                adjacent_searcher = ElementSearcher(
                    primitive.TAG,
                    self._adjacent_resource_id,
                    self._group_element,
                )
                if adjacent_searcher.element_found():
                    self._adjacent_resource_element = (
                        adjacent_searcher.get_element()
                    )
                else:
                    report_list.append(
                        reports
                        .cannot_group_resource_adjacent_resource_not_in_group(
                            self._adjacent_resource_id,
                            self._group_id,
                        )
                    )
            # The group will be created so there is no adjacent resource in it.
            elif group_missing_id_valid:
                report_list.append(
                    reports
                    .cannot_group_resource_adjacent_resource_for_new_group(
                        self._adjacent_resource_id,
                        self._group_id,
                    )
                )
            # else: The group_id belongs to a non-group element, checking the
            # adjacent_reource is pointless.

        report_list.extend(
            self._validate_elements(
                bad_or_missing_group_specified=(
                    self._group_element is None
                ),
                bad_resources_specified=(
                    self._resource_id_list
                    and
                    not self._resource_element_list
                ),
                bad_adjacent_specified=(
                    self._adjacent_resource_id
                    and
                    self._adjacent_resource_element is None
                )
            )
        )

        return report_list

    def group_element(self):
        return self._group_element

    def resource_element_list(self):
        return self._resource_element_list

    def adjacent_resource_element(self):
        return self._adjacent_resource_element


def move_resources_to_group(
    group_element, primitives_to_place, adjacent_resource=None,
    put_after_adjacent=True
):
    """
    Put resources into a group or move them within their group

    etree.Element group_element -- the group to put resources into
    iterable primitives_to_place -- resource elements to put into the group
    etree.Element adjacent_resource -- put resources beside this one if set
    bool put_after_adjacent -- put resources after or before the adjacent one
    """
    for resource in primitives_to_place:
        old_parent = resource.getparent()

        # Move a resource to the group.
        if (
            adjacent_resource is not None
            and
            adjacent_resource.getnext() is not None
            and
            put_after_adjacent
        ):
            adjacent_resource.getnext().addprevious(resource)
            adjacent_resource = resource
        elif (
            adjacent_resource is not None
            and
            not put_after_adjacent
        ):
            adjacent_resource.addprevious(resource)
        else:
            group_element.append(resource)
            adjacent_resource = resource

        # If the resource was the last resource in another group, that group is
        # now empty and must be deleted. If the group is in a clone element,
        # delete that as well.
        if (
            old_parent is not None
            and
            group.is_group(old_parent)
            and
            not group.get_inner_resources(old_parent)
            and
            old_parent.getparent() is not None
        ):
            old_grandparent = old_parent.getparent()
            if (
                clone.is_any_clone(old_grandparent)
                and
                old_grandparent.getparent() is not None
            ):
                old_grandparent.getparent().remove(old_grandparent)
            else:
                old_grandparent.remove(old_parent)



# pylint: disable=too-many-lines
import math
import os.path
import time

from pcs import settings
from pcs.common import (
    file_type_codes,
    report_codes,
    ssl,
)
from pcs.common.file import RawFileError
from pcs.common.node_communicator import HostNotFound
from pcs.common.tools import (
    format_environment_error,
    join_multilines,
)
from pcs.common.reports import SimpleReportProcessor
from pcs.lib import reports, node_communication_format, sbd, validate
from pcs.lib.booth import sync as booth_sync
from pcs.lib.cib import fencing_topology
from pcs.lib.cib.resource.remote_node import find_node_list as get_remote_nodes
from pcs.lib.cib.resource.guest_node import find_node_list as get_guest_nodes
from pcs.lib.cib.tools import (
    get_fencing_topology,
    get_resources,
)
from pcs.lib.communication import cluster
from pcs.lib.communication.corosync import (
    CheckCorosyncOffline,
    DistributeCorosyncConf,
    ReloadCorosyncConf,
)
from pcs.lib.communication.nodes import (
    CheckPacemakerStarted,
    DistributeFilesWithoutForces,
    EnableCluster,
    GetHostInfo,
    GetOnlineTargets,
    RemoveFilesWithoutForces,
    RemoveNodesFromCib,
    SendPcsdSslCertAndKey,
    StartCluster,
    UpdateKnownHosts,
)
from pcs.lib.communication.sbd import (
    CheckSbd,
    SetSbdConfig,
    EnableSbdService,
    DisableSbdService,
)
from pcs.lib.communication.tools import (
    run as run_com,
    run_and_raise,
)
from pcs.lib.corosync import (
    config_facade,
    config_parser,
    config_validators,
    constants as corosync_constants,
    qdevice_net,
)
from pcs.lib.node import get_existing_nodes_names
from pcs.lib.errors import (
    LibraryError,
    ReportItemSeverity,
)
from pcs.lib.external import is_service_running
from pcs.lib.pacemaker.live import (
    get_cib,
    get_cib_xml,
    get_cib_xml_cmd_results,
    get_cluster_status_xml,
    remove_node,
    verify as verify_cmd,
)
from pcs.lib.pacemaker.state import ClusterState
from pcs.lib.pacemaker.values import get_valid_timeout_seconds
from pcs.lib.tools import (
    environment_file_to_dict,
    generate_binary_key,
)


def node_clear(env, node_name, allow_clear_cluster_node=False):
    """
    Remove specified node from various cluster caches.

    LibraryEnvironment env provides all for communication with externals
    string node_name
    bool allow_clear_cluster_node -- flag allows to clear node even if it's
        still in a cluster
    """
    _ensure_live_env(env) # raises if env is not live

    current_nodes, report_list = get_existing_nodes_names(
        env.get_corosync_conf(),
        env.get_cib()
    )
    env.report_processor.process_list(report_list)

    if node_name in current_nodes:
        env.report_processor.process(
            reports.get_problem_creator(
                report_codes.FORCE_CLEAR_CLUSTER_NODE,
                allow_clear_cluster_node
            )(
                reports.node_to_clear_is_still_in_cluster,
                node_name
            )
        )

    remove_node(env.cmd_runner(), node_name)

def verify(env, verbose=False):
    runner = env.cmd_runner()
    dummy_stdout, verify_stderr, verify_returncode, can_be_more_verbose = (
        verify_cmd(runner, verbose=verbose)
    )

    #1) Do not even try to think about upgrading!
    #2) We do not need cib management in env (no need for push...).
    #So env.get_cib is not best choice here (there were considerations to
    #upgrade cib at all times inside env.get_cib). Go to a lower level here.
    if verify_returncode != 0:
        env.report_processor.append(
            reports.invalid_cib_content(verify_stderr, can_be_more_verbose)
        )

        #Cib is sometimes loadable even if `crm_verify` fails (e.g. when
        #fencing topology is invalid). On the other hand cib with id duplication
        #is not loadable.
        #We try extra checks when cib is possible to load.
        cib_xml, dummy_stderr, returncode = get_cib_xml_cmd_results(runner)
        if returncode != 0:
            #can raise; raise LibraryError is better but in this case we prefer
            #be consistent with raising below
            env.report_processor.send()
    else:
        cib_xml = get_cib_xml(runner)

    cib = get_cib(cib_xml)
    fencing_topology.verify(
        env.report_processor,
        get_fencing_topology(cib),
        get_resources(cib),
        ClusterState(get_cluster_status_xml(runner)).node_section.nodes
    )
    #can raise
    env.report_processor.send()

def setup(
    env, cluster_name, nodes,
    transport_type=None, transport_options=None, link_list=None,
    compression_options=None, crypto_options=None, totem_options=None,
    quorum_options=None, wait=False, start=False, enable=False,
    no_keys_sync=False, force_flags=None,
):
    # pylint: disable=too-many-arguments, too-many-locals, too-many-statements
    """
    Set up cluster on specified nodes.
    Validation of the inputs is done here. Possible existing clusters are
    destroyed (when using force). Authkey files for corosync and pacemaer,
    known hosts and and newly generated corosync.conf are distributed to all
    nodes.
    Raise LibraryError on any error.

    env LibraryEnvironment
    cluster_name string -- name of a cluster to set up
    nodes list -- list of dicts which represents node.
        Supported keys are: name (required), addrs. See note bellow.
    transport_type string -- transport type of a cluster
    transport_options dict -- transport specific options
    link_list list of dict -- list of links, depends of transport_type
    compression_options dict -- only available for knet transport. In
        corosync.conf they are prefixed 'knet_compression_'
    crypto_options dict -- only available for knet transport'. In corosync.conf
        they are prefixed 'crypto_'
    totem_options dict -- options of section 'totem' in corosync.conf
    quorum_options dict -- options of section 'quorum' in corosync.conf
    wait -- specifies if command should try to wait for cluster to start up.
        Has no effect start is False. If set to False command will not wait for
        cluster to start. If None command will wait for some default timeout.
        If int wait set timeout to int value of seconds.
    start bool -- if True start cluster when it is set up
    enable bool -- if True enable cluster when it is set up
    no_keys_sync bool -- if True do not crete and distribute files: pcsd ssl
        cert and key, pacemaker authkey, corosync authkey
    force_flags list -- list of flags codes

    The command is defaulting node addresses if they are not specified. The
    defaulting is done for each node individually if and only if the "addrs" key
    is not present for the node. If the "addrs" key is present and holds an
    empty list, no defaulting is done.
    This will default addresses for node2 and won't modify addresses for other
    nodes (no addresses will be defined for node3):
    nodes=[
        {"name": "node1", "addrs": ["node1-addr"]},
        {"name": "node2"},
        {"name": "node3", "addrs": []},
    ]
    """
    _ensure_live_env(env) # raises if env is not live
    if force_flags is None:
        force_flags = []
    force = report_codes.FORCE in force_flags

    transport_type = transport_type or "knet"
    transport_options = transport_options or {}
    link_list = link_list or []
    compression_options = compression_options or {}
    crypto_options = crypto_options or {}
    totem_options = totem_options or {}
    quorum_options = quorum_options or {}
    nodes = [
        _normalize_dict(node, {"addrs"}) for node in nodes
    ]
    if (
        transport_type in corosync_constants.TRANSPORTS_KNET
        and
        not crypto_options
    ):
        crypto_options = {
            "cipher": "aes256",
            "hash": "sha256",
        }
    # Get IP version for node addresses validation. Defaults taken from man
    # corosync.conf
    ip_version = (
        corosync_constants.IP_VERSION_4 if transport_type == "udp"
        else corosync_constants.IP_VERSION_64
    )
    if (
        transport_options.get("ip_version")
        in
        corosync_constants.IP_VERSION_VALUES
    ):
        ip_version = transport_options["ip_version"]

    report_processor = SimpleReportProcessor(env.report_processor)
    target_factory = env.get_node_target_factory()

    # Get targets for all nodes and report unknown (== not-authorized) nodes.
    # If a node doesn't contain the 'name' key, validation of inputs reports it.
    # That means we don't report missing names but cannot rely on them being
    # present either.
    target_report_list, target_list = (
        target_factory.get_target_list_with_reports(
            [node["name"] for node in nodes if "name" in node],
            allow_skip=False,
        )
    )
    report_processor.report_list(target_report_list)

    # Use an address defined in known-hosts for each node with no addresses
    # specified. This allows users not to specify node addresses at all which
    # simplifies the whole cluster setup command / form significantly.
    addrs_defaulter = _get_addrs_defaulter(
        report_processor,
        {target.label: target for target in target_list}
    )
    nodes = [
        _set_defaults_in_dict(node, {"addrs": addrs_defaulter})
        for node in nodes
    ]

    # Validate inputs.
    report_processor.report_list(config_validators.create(
        cluster_name, nodes, transport_type, ip_version,
        force_unresolvable=force
    ))
    max_node_addr_count = max(
        [len(node["addrs"]) for node in nodes],
        default=0
    )
    if transport_type in corosync_constants.TRANSPORTS_KNET:
        report_processor.report_list(
            config_validators.create_transport_knet(
                transport_options,
                compression_options,
                crypto_options
            )
            +
            config_validators.create_link_list_knet(
                link_list,
                max_node_addr_count
            )
        )
    elif transport_type in corosync_constants.TRANSPORTS_UDP:
        report_processor.report_list(
            config_validators.create_transport_udp(
                transport_options,
                compression_options,
                crypto_options
            )
            +
            config_validators.create_link_list_udp(
                link_list,
                max_node_addr_count
            )
        )
    report_processor.report_list(
        config_validators.create_totem(totem_options)
        +
        # We are creating the config and we know there is no qdevice in it.
        config_validators.create_quorum_options(quorum_options, False)
    )

    # Validate flags
    wait_timeout = _get_validated_wait_timeout(report_processor, wait, start)

    # Validate the nodes
    com_cmd = GetHostInfo(report_processor)
    com_cmd.set_targets(target_list)
    report_processor.report_list(
        _host_check_cluster_setup(
            run_com(env.get_node_communicator(), com_cmd), force
        )
    )

    # If there is an error reading the file, this will report it and exit
    # safely before any change is made to the nodes.
    sync_ssl_certs = _is_ssl_cert_sync_enabled(report_processor)

    if report_processor.has_errors:
        raise LibraryError()

    # Validation done. If errors occured, an exception has been raised and we
    # don't get below this line.

    # Destroy cluster on all nodes.
    com_cmd = cluster.Destroy(env.report_processor)
    com_cmd.set_targets(target_list)
    run_and_raise(env.get_node_communicator(), com_cmd)

    # Distribute auth tokens.
    com_cmd = UpdateKnownHosts(
        env.report_processor,
        known_hosts_to_add=env.get_known_hosts(
            [target.label for target in target_list]
        ),
        known_hosts_to_remove=[],
    )
    com_cmd.set_targets(target_list)
    run_and_raise(env.get_node_communicator(), com_cmd)

    # TODO This should be in the file distribution call but so far we don't
    # have a call which allows to save and delete files at the same time.
    com_cmd = RemoveFilesWithoutForces(
        env.report_processor, {"pcsd settings": {"type": "pcsd_settings"}},
    )
    com_cmd.set_targets(target_list)
    run_and_raise(env.get_node_communicator(), com_cmd)

    if not no_keys_sync:
        # Distribute configuration files except corosync.conf. Sending
        # corosync.conf serves as a "commit" as its presence on a node marks the
        # node as a part of a cluster.
        corosync_authkey = generate_binary_key(
            random_bytes_count=settings.corosync_authkey_bytes
        )
        pcmk_authkey = generate_binary_key(
            random_bytes_count=settings.pacemaker_authkey_bytes
        )
        actions = {}
        actions.update(
            node_communication_format.corosync_authkey_file(corosync_authkey)
        )
        actions.update(
            node_communication_format.pcmk_authkey_file(pcmk_authkey)
        )
        com_cmd = DistributeFilesWithoutForces(env.report_processor, actions)
        com_cmd.set_targets(target_list)
        run_and_raise(env.get_node_communicator(), com_cmd)

        # Distribute and reload pcsd SSL certificate
        if sync_ssl_certs:
            report_processor.report(
                reports.pcsd_ssl_cert_and_key_distribution_started(
                    [target.label for target in target_list]
                )
            )
            # Local certificate and key cannot be used because the local node
            # may not be a part of the new cluter at all.
            ssl_key_raw = ssl.generate_key()
            ssl_key = ssl.dump_key(ssl_key_raw)
            ssl_cert = ssl.dump_cert(
                ssl.generate_cert(ssl_key_raw, target_list[0].label)
            )
            com_cmd = SendPcsdSslCertAndKey(
                env.report_processor,
                ssl_cert, ssl_key
            )
            com_cmd.set_targets(target_list)
            run_and_raise(env.get_node_communicator(), com_cmd)

    # Create and distribute corosync.conf. Once a node saves corosync.conf it
    # is considered to be in a cluster.
    corosync_conf = config_facade.ConfigFacade.create(
        cluster_name, nodes, transport_type
    )
    corosync_conf.set_totem_options(totem_options)
    corosync_conf.set_quorum_options(quorum_options)
    corosync_conf.create_link_list(link_list)
    if transport_type in corosync_constants.TRANSPORTS_KNET:
        corosync_conf.set_transport_knet_options(
            transport_options, compression_options, crypto_options
        )
    elif transport_type in corosync_constants.TRANSPORTS_UDP:
        corosync_conf.set_transport_udp_options(transport_options)

    _verify_corosync_conf(corosync_conf) # raises if corosync not valid
    com_cmd = DistributeFilesWithoutForces(
        env.report_processor,
        node_communication_format.corosync_conf_file(
            corosync_conf.config.export()
        ),
    )
    com_cmd.set_targets(target_list)
    run_and_raise(env.get_node_communicator(), com_cmd)

    env.report_processor.process(reports.cluster_setup_success())

    # Optionally enable and start cluster services.
    if enable:
        com_cmd = EnableCluster(env.report_processor)
        com_cmd.set_targets(target_list)
        run_and_raise(env.get_node_communicator(), com_cmd)
    if start:
        _start_cluster(
            env.communicator_factory,
            env.report_processor,
            target_list,
            wait_timeout=wait_timeout,
        )

def add_nodes(
    env, nodes, wait=False, start=False, enable=False,
    no_watchdog_validation=False, force_flags=None,
):
    # pylint: disable=too-many-locals, too-many-statements, too-many-branches
    """
    Add specified nodes to the local cluster
    Raise LibraryError on any error.

    env LibraryEnvironment
    nodes list -- list of dicts which represents node.
        Supported keys are: name (required), addrs (list), devices (list),
        watchdog. See note below.
    wait -- specifies if command should try to wait for cluster to start up.
        Has no effect start is False. If set to False command will not wait for
        cluster to start. If None command will wait for some default timeout.
        If int wait set timeout to int value of seconds.
    start bool -- if True start cluster when it is set up
    enable bool -- if True enable cluster when it is set up
    no_watchdog_validation bool -- if True do not validate specified watchdogs
        on remote hosts
    force_flags list -- list of flags codes

    The command is defaulting node addresses if they are not specified. The
    defaulting is done for each node individually if and only if the "addrs" key
    is not present for the node. If the "addrs" key is present and holds an
    empty list, no defaulting is done.
    This will default addresses for node2 and won't modify addresses for other
    nodes (no addresses will be defined for node3):
    nodes=[
        {"name": "node1", "addrs": ["node1-addr"]},
        {"name": "node2"},
        {"name": "node3", "addrs": []},
    ]
    """
    _ensure_live_env(env) # raises if env is not live

    if force_flags is None:
        force_flags = []
    force = report_codes.FORCE in force_flags
    skip_offline_nodes = report_codes.SKIP_OFFLINE_NODES in force_flags

    report_processor = SimpleReportProcessor(env.report_processor)
    target_factory = env.get_node_target_factory()
    is_sbd_enabled = sbd.is_sbd_enabled(env.cmd_runner())
    corosync_conf = env.get_corosync_conf()
    corosync_node_options = {"name", "addrs"}
    sbd_node_options = {"devices", "watchdog"}

    keys_to_normalize = {"addrs"}
    if is_sbd_enabled:
        keys_to_normalize |= sbd_node_options
    new_nodes = [_normalize_dict(node, keys_to_normalize) for node in nodes]

    # get targets for existing nodes
    cluster_nodes_names, nodes_report_list = get_existing_nodes_names(
        corosync_conf,
        # Pcs is unable to communicate with nodes missing names. It cannot send
        # new corosync.conf to them. That might break the cluster. Hence we
        # error out.
        error_on_missing_name=True
    )
    report_processor.report_list(nodes_report_list)
    target_report_list, cluster_nodes_target_list = (
        target_factory.get_target_list_with_reports(
            cluster_nodes_names,
            skip_non_existing=skip_offline_nodes,
        )
    )
    report_processor.report_list(target_report_list)
    # get a target for qnetd if needed
    qdevice_model, qdevice_model_options, _, _ = (
        corosync_conf.get_quorum_device_settings()
    )
    if qdevice_model == "net":
        try:
            qnetd_target = target_factory.get_target(
                qdevice_model_options["host"]
            )
        except HostNotFound:
            report_processor.report(
                reports.host_not_found([qdevice_model_options["host"]])
            )

    # Get targets for new nodes and report unknown (== not-authorized) nodes.
    # If a node doesn't contain the 'name' key, validation of inputs reports it.
    # That means we don't report missing names but cannot rely on them being
    # present either.
    target_report_list, new_nodes_target_list = (
        target_factory.get_target_list_with_reports(
            [node["name"] for node in new_nodes if "name" in node],
            allow_skip=False,
        )
    )
    report_processor.report_list(target_report_list)

    # Set default values for not-specified node options.
    # Use an address defined in known-hosts for each node with no addresses
    # specified. This allows users not to specify node addresses at all which
    # simplifies the whole node add command / form significantly.
    new_nodes_target_dict = {
        target.label: target for target in new_nodes_target_list
    }
    addrs_defaulter = _get_addrs_defaulter(
        report_processor,
        new_nodes_target_dict
    )
    new_nodes_defaulters = {"addrs": addrs_defaulter}
    if is_sbd_enabled:
        watchdog_defaulter = _get_watchdog_defaulter(
            report_processor,
            new_nodes_target_dict
        )
        new_nodes_defaulters["devices"] = lambda _: []
        new_nodes_defaulters["watchdog"] = watchdog_defaulter
    new_nodes = [
        _set_defaults_in_dict(node, new_nodes_defaulters) for node in new_nodes
    ]
    new_nodes_dict = {
        node["name"]: node for node in new_nodes if "name" in node
    }

    # Validate inputs - node options names
    # We do not want to make corosync validators know about SBD options and
    # vice versa. Therefore corosync and SBD validators get only valid corosync
    # and SBD options respectively, and we need to check for any surplus
    # options here.
    report_processor.report_list(
        validate.NamesIn(
            corosync_node_options | sbd_node_options,
            option_type="node"
        ).validate({
            # Get a dict containing options of all nodes. Values don't matter
            # for validate.NamesIn validator.
            option_name: None
            for node_option_names in [node.keys() for node in new_nodes]
            for option_name in node_option_names
        })
    )

    # Validate inputs - corosync part
    try:
        cib = env.get_cib()
        cib_nodes = get_remote_nodes(cib) + get_guest_nodes(cib)
    except LibraryError:
        cib_nodes = []
        report_processor.report(
            reports.get_problem_creator(
                report_codes.FORCE_LOAD_NODES_FROM_CIB,
                force
            )(
                reports.cib_load_error_get_nodes_for_validation
            )
        )
    # corosync validator rejects non-corosync keys
    new_nodes_corosync = [
        {key: node[key] for key in corosync_node_options if key in node}
        for node in new_nodes
    ]
    report_processor.report_list(config_validators.add_nodes(
        new_nodes_corosync,
        corosync_conf.get_nodes(),
        cib_nodes,
        force_unresolvable=force
    ))

    # Validate inputs - SBD part
    if is_sbd_enabled:
        report_processor.report_list(
            sbd.validate_new_nodes_devices(
                {
                    node["name"]: node["devices"]
                    for node in new_nodes if "name" in node
                }
            )
        )
    else:
        for node in new_nodes:
            sbd_options = sbd_node_options.intersection(node.keys())
            if sbd_options and "name" in node:
                report_processor.report(
                    reports.sbd_not_used_cannot_set_sbd_options(
                        sbd_options,
                        node["name"]
                    )
                )

    # Validate inputs - flags part
    wait_timeout = _get_validated_wait_timeout(report_processor, wait, start)

    # Get online cluster nodes
    # This is the only call in which we accept skip_offline_nodes option for the
    # cluster nodes. In all the other actions we communicate only with the
    # online nodes. This allows us to simplify code as any communication issue
    # is considered an error, ends the command processing and is not possible
    # to skip it by skip_offline_nodes. We do not have to care about a situation
    # when a communication command cannot connect to some nodes and then the
    # next command can connect but fails due to the previous one did not
    # succeed.
    online_cluster_target_list = []
    if cluster_nodes_target_list:
        com_cmd = GetOnlineTargets(
            report_processor, ignore_offline_targets=skip_offline_nodes,
        )
        com_cmd.set_targets(cluster_nodes_target_list)
        online_cluster_target_list = run_com(
            env.get_node_communicator(), com_cmd
        )
        offline_cluster_target_list = [
            target for target in cluster_nodes_target_list
            if target not in online_cluster_target_list
        ]
        if not online_cluster_target_list:
            report_processor.report(
                reports.unable_to_perform_operation_on_any_node()
            )
        elif offline_cluster_target_list and skip_offline_nodes:
            # TODO: report (warn) how to fix offline nodes when they come online
            # report_processor.report(None)
            pass

    # Validate existing cluster nodes status
    atb_has_to_be_enabled = sbd.atb_has_to_be_enabled(
        env.cmd_runner(), corosync_conf, len(new_nodes)
    )
    if atb_has_to_be_enabled:
        report_processor.report(
            reports.corosync_quorum_atb_will_be_enabled_due_to_sbd()
        )
        if online_cluster_target_list:
            com_cmd = CheckCorosyncOffline(
                report_processor, allow_skip_offline=False,
            )
            com_cmd.set_targets(online_cluster_target_list)
            run_com(env.get_node_communicator(), com_cmd)

    # Validate new nodes. All new nodes have to be online.
    com_cmd = GetHostInfo(report_processor)
    com_cmd.set_targets(new_nodes_target_list)
    report_processor.report_list(
        _host_check_cluster_setup(
            run_com(env.get_node_communicator(), com_cmd),
            force,
            # version of services may not be the same across the existing
            # cluster nodes, so it's not easy to make this check properly
            check_services_versions=False,
        )
    )

    # Validate SBD on new nodes
    if is_sbd_enabled:
        if no_watchdog_validation:
            report_processor.report(reports.sbd_watchdog_validation_inactive())
        com_cmd = CheckSbd(report_processor)
        for new_node_target in new_nodes_target_list:
            new_node = new_nodes_dict[new_node_target.label]
            # Do not send watchdog if validation is turned off. Listing of
            # available watchdogs in pcsd may restart the machine in some
            # corner cases.
            com_cmd.add_request(
                new_node_target,
                watchdog="" if no_watchdog_validation else new_node["watchdog"],
                device_list=new_node["devices"],
            )
        run_com(env.get_node_communicator(), com_cmd)

    # If there is an error reading the file, this will report it and exit
    # safely before any change is made to the nodes.
    sync_ssl_certs = _is_ssl_cert_sync_enabled(report_processor)

    if report_processor.has_errors:
        raise LibraryError()

    # Validation done. If errors occured, an exception has been raised and we
    # don't get below this line.

    # First set up everything else than corosync. Once the new nodes are present
    # in corosync.conf, they're considered part of a cluster and the node add
    # command cannot be run again. So we need to minimize the amout of actions
    # (and therefore possible failures) after adding the nodes to corosync.

    # distribute auth tokens of all cluster nodes (including the new ones) to
    # all new nodes
    com_cmd = UpdateKnownHosts(
        env.report_processor,
        known_hosts_to_add=env.get_known_hosts(
            cluster_nodes_names + list(new_nodes_dict.keys())
        ),
        known_hosts_to_remove=[],
    )
    com_cmd.set_targets(new_nodes_target_list)
    run_and_raise(env.get_node_communicator(), com_cmd)

    # qdevice setup
    if qdevice_model == "net":
        qdevice_net.set_up_client_certificates(
            env.cmd_runner(),
            env.report_processor,
            env.communicator_factory,
            qnetd_target,
            corosync_conf.get_cluster_name(),
            new_nodes_target_list,
            # we don't want to allow skiping offline nodes which are being
            # added, otherwise qdevice will not work properly
            skip_offline_nodes=False,
            allow_skip_offline=False
        )

    # sbd setup
    if is_sbd_enabled:
        sbd_cfg = environment_file_to_dict(sbd.get_local_sbd_config())

        com_cmd = SetSbdConfig(env.report_processor)
        for new_node_target in new_nodes_target_list:
            new_node = new_nodes_dict[new_node_target.label]
            com_cmd.add_request(
                new_node_target,
                sbd.create_sbd_config(
                    sbd_cfg,
                    new_node["name"],
                    watchdog=new_node["watchdog"],
                    device_list=new_node["devices"],
                )
            )
        run_and_raise(env.get_node_communicator(), com_cmd)

        com_cmd = EnableSbdService(env.report_processor)
        com_cmd.set_targets(new_nodes_target_list)
        run_and_raise(env.get_node_communicator(), com_cmd)
    else:
        com_cmd = DisableSbdService(env.report_processor)
        com_cmd.set_targets(new_nodes_target_list)
        run_and_raise(env.get_node_communicator(), com_cmd)

    # booth setup
    booth_sync.send_all_config_to_node(
        env.get_node_communicator(),
        env.report_processor,
        new_nodes_target_list,
        rewrite_existing=force,
        skip_wrong_config=force,
    )

    # distribute corosync and pacemaker authkeys
    files_action = {}
    forceable_io_error_creator = reports.get_problem_creator(
        report_codes.SKIP_FILE_DISTRIBUTION_ERRORS, force
    )
    if os.path.isfile(settings.corosync_authkey_file):
        try:
            files_action.update(
                node_communication_format.corosync_authkey_file(
                    open(settings.corosync_authkey_file, "rb").read()
                )
            )
        except EnvironmentError as e:
            report_processor.report(forceable_io_error_creator(
                reports.file_io_error,
                file_type_codes.COROSYNC_AUTHKEY,
                RawFileError.ACTION_READ,
                format_environment_error(e),
                file_path=settings.corosync_authkey_file,
            ))

    if os.path.isfile(settings.pacemaker_authkey_file):
        try:
            files_action.update(
                node_communication_format.pcmk_authkey_file(
                    open(settings.pacemaker_authkey_file, "rb").read()
                )
            )
        except EnvironmentError as e:
            report_processor.report(forceable_io_error_creator(
                reports.file_io_error,
                file_type_codes.PACEMAKER_AUTHKEY,
                RawFileError.ACTION_READ,
                format_environment_error(e),
                file_path=settings.pacemaker_authkey_file,
            ))

    # pcs_settings.conf was previously synced using pcsdcli send_local_configs.
    # This has been changed temporarily until new system for distribution and
    # syncronization of configs will be introduced.
    if os.path.isfile(settings.pcsd_settings_conf_location):
        try:
            files_action.update(
                node_communication_format.pcs_settings_conf_file(
                    open(settings.pcsd_settings_conf_location, "r").read()
                )
            )
        except EnvironmentError as e:
            report_processor.report(forceable_io_error_creator(
                reports.file_io_error,
                file_type_codes.PCS_SETTINGS_CONF,
                RawFileError.ACTION_READ,
                format_environment_error(e),
                file_path=settings.pcsd_settings_conf_location,
            ))

    # stop here if one of the files could not be loaded and it was not forced
    if report_processor.has_errors:
        raise LibraryError()

    if files_action:
        com_cmd = DistributeFilesWithoutForces(
            env.report_processor,
            files_action
        )
        com_cmd.set_targets(new_nodes_target_list)
        run_and_raise(env.get_node_communicator(), com_cmd)

    # Distribute and reload pcsd SSL certificate
    if sync_ssl_certs:
        report_processor.report(
            reports.pcsd_ssl_cert_and_key_distribution_started(
                [target.label for target in new_nodes_target_list]
            )
        )

        try:
            with open(settings.pcsd_cert_location, "r") as file:
                ssl_cert = file.read()
        except EnvironmentError as e:
            report_processor.report(
                reports.file_io_error(
                    file_type_codes.PCSD_SSL_CERT,
                    RawFileError.ACTION_READ,
                    format_environment_error(e),
                    file_path=settings.pcsd_cert_location,
                )
            )
        try:
            with open(settings.pcsd_key_location, "r") as file:
                ssl_key = file.read()
        except EnvironmentError as e:
            report_processor.report(
                reports.file_io_error(
                    file_type_codes.PCSD_SSL_KEY,
                    RawFileError.ACTION_READ,
                    format_environment_error(e),
                    file_path=settings.pcsd_key_location,
                )
            )
        if report_processor.has_errors:
            raise LibraryError()

        com_cmd = SendPcsdSslCertAndKey(env.report_processor, ssl_cert, ssl_key)
        com_cmd.set_targets(new_nodes_target_list)
        run_and_raise(env.get_node_communicator(), com_cmd)

    # When corosync >= 2 is in use, the procedure for adding a node is:
    # 1. add the new node to corosync.conf on all existing nodes
    # 2. reload corosync.conf before the new node is started
    # 3. start the new node
    # If done otherwise, membership gets broken and qdevice hangs. Cluster
    # will recover after a minute or so but still it's a wrong way.

    corosync_conf.add_nodes(new_nodes_corosync)
    if atb_has_to_be_enabled:
        corosync_conf.set_quorum_options(dict(auto_tie_breaker="1"))

    _verify_corosync_conf(corosync_conf) # raises if corosync not valid
    com_cmd = DistributeCorosyncConf(
        env.report_processor,
        corosync_conf.config.export(),
        allow_skip_offline=False,
    )
    com_cmd.set_targets(online_cluster_target_list + new_nodes_target_list)
    run_and_raise(env.get_node_communicator(), com_cmd)

    com_cmd = ReloadCorosyncConf(env.report_processor)
    com_cmd.set_targets(online_cluster_target_list)
    run_and_raise(env.get_node_communicator(), com_cmd)

    # Optionally enable and start cluster services.
    if enable:
        com_cmd = EnableCluster(env.report_processor)
        com_cmd.set_targets(new_nodes_target_list)
        run_and_raise(env.get_node_communicator(), com_cmd)
    if start:
        _start_cluster(
            env.communicator_factory,
            env.report_processor,
            new_nodes_target_list,
            wait_timeout=wait_timeout,
        )

def _ensure_live_env(env):
    not_live = []
    if not env.is_cib_live:
        not_live.append("CIB")
    if not env.is_corosync_conf_live:
        not_live.append("COROSYNC_CONF")
    if not_live:
        raise LibraryError(reports.live_environment_required(not_live))

def _start_cluster(
    communicator_factory, report_processor, target_list, wait_timeout=False,
):
    # Large clusters take longer time to start up. So we make the timeout
    # longer for each 8 nodes:
    #  1 -  8 nodes: 1 * timeout
    #  9 - 16 nodes: 2 * timeout
    # 17 - 24 nodes: 3 * timeout
    # and so on ...
    # Users can override this and set their own timeout by specifying
    # the --request-timeout option.
    timeout = int(
        settings.default_request_timeout * math.ceil(len(target_list) / 8.0)
    )
    com_cmd = StartCluster(report_processor)
    com_cmd.set_targets(target_list)
    run_and_raise(
        communicator_factory.get_communicator(request_timeout=timeout), com_cmd
    )
    if wait_timeout is not False:
        report_processor.process_list(
            _wait_for_pacemaker_to_start(
                communicator_factory.get_communicator(),
                report_processor,
                target_list,
                timeout=wait_timeout, # wait_timeout is either None or a timeout
            )
        )

def _wait_for_pacemaker_to_start(
    node_communicator, report_processor, target_list, timeout=None
):
    timeout = 60 * 15 if timeout is None else timeout
    interval = 2
    stop_at = time.time() + timeout
    report_processor.process(
        reports.wait_for_node_startup_started(
            [target.label for target in target_list]
        )
    )
    error_report_list = []
    while target_list:
        if time.time() > stop_at:
            error_report_list.append(reports.wait_for_node_startup_timed_out())
            break
        time.sleep(interval)
        com_cmd = CheckPacemakerStarted(report_processor)
        com_cmd.set_targets(target_list)
        target_list = run_com(node_communicator, com_cmd)
        error_report_list.extend(com_cmd.error_list)

    if error_report_list:
        error_report_list.append(reports.wait_for_node_startup_error())
    return error_report_list

def _host_check_cluster_setup(
    host_info_dict, force, check_services_versions=True
):
    # pylint: disable=too-many-locals
    report_list = []
    # We only care about services which matter for creating a cluster. It does
    # not make sense to check e.g. booth when a) it will never be used b) it
    # will be used in a year - which means we should do the check in a year.
    service_version_dict = {
        "pacemaker": {},
        "corosync": {},
        "pcsd": {},
    }
    required_service_list = ["pacemaker", "corosync"]
    required_as_stopped_service_list = (
        required_service_list + ["pacemaker_remote"]
    )
    report_severity = (
        ReportItemSeverity.WARNING if force else ReportItemSeverity.ERROR
    )
    report_forceable = (
        None if force else report_codes.FORCE_ALREADY_IN_CLUSTER
    )
    cluster_exists_on_nodes = False
    for host_name, host_info in host_info_dict.items():
        try:
            services = host_info["services"]
            if check_services_versions:
                for service, version_dict in service_version_dict.items():
                    version_dict[host_name] = services[service]["version"]
            missing_service_list = [
                service for service in required_service_list
                if not services[service]["installed"]
            ]
            if missing_service_list:
                report_list.append(reports.service_not_installed(
                    host_name, missing_service_list
                ))
            cannot_be_running_service_list = [
                service for service in required_as_stopped_service_list
                if service in services and services[service]["running"]
            ]
            if cannot_be_running_service_list:
                cluster_exists_on_nodes = True
                report_list.append(
                    reports.host_already_in_cluster_services(
                        host_name,
                        cannot_be_running_service_list,
                        severity=report_severity,
                        forceable=report_forceable,
                    )
                )
            if host_info["cluster_configuration_exists"]:
                cluster_exists_on_nodes = True
                report_list.append(
                    reports.host_already_in_cluster_config(
                        host_name,
                        severity=report_severity,
                        forceable=report_forceable,
                    )
                )
        except KeyError:
            report_list.append(reports.invalid_response_format(host_name))

    if check_services_versions:
        for service, version_dict in service_version_dict.items():
            report_list.extend(
                _check_for_not_matching_service_versions(service, version_dict)
            )

    if cluster_exists_on_nodes and not force:
        # This is always a forceable error
        report_list.append(reports.cluster_will_be_destroyed())
    return report_list

def _check_for_not_matching_service_versions(service, service_version_dict):
    if len(set(service_version_dict.values())) <= 1:
        return []
    return [
        reports.service_version_mismatch(service, service_version_dict)
    ]

def _normalize_dict(input_dict, required_keys):
    normalized = dict(input_dict)
    for key in required_keys:
        if key not in normalized:
            normalized[key] = None
    return normalized

def _set_defaults_in_dict(input_dict, defaults):
    completed = dict(input_dict)
    for key, factory in defaults.items():
        if completed[key] is None:
            completed[key] = factory(input_dict)
    return completed

def _get_addrs_defaulter(
    report_processor: SimpleReportProcessor, targets_dict
):
    def defaulter(node):
        if "name" not in node:
            return []
        target = targets_dict.get(node["name"])
        if target:
            report_processor.report(
                reports.using_known_host_address_for_host(
                    node["name"], target.first_addr
                )
            )
            return [target.first_addr]
        return []
    return defaulter

def _get_watchdog_defaulter(
    report_processor: SimpleReportProcessor, targets_dict
):
    # pylint: disable=unused-argument
    def defaulter(node):
        report_processor.report(reports.using_default_watchdog(
            settings.sbd_watchdog_default,
            node["name"],
        ))
        return settings.sbd_watchdog_default
    return defaulter

def _get_validated_wait_timeout(report_processor, wait, start):
    try:
        if wait is False:
            return False
        if not start:
            report_processor.report(
                reports.wait_for_node_startup_without_start()
            )
        return get_valid_timeout_seconds(wait)
    except LibraryError as e:
        report_processor.report_list(e.args)
    return None

def _is_ssl_cert_sync_enabled(report_processor):
    try:
        if os.path.isfile(settings.pcsd_config):
            with open(settings.pcsd_config, "r") as cfg_file:
                cfg = environment_file_to_dict(cfg_file.read())
                return (
                    cfg.get("PCSD_SSL_CERT_SYNC_ENABLED", "false").lower()
                    ==
                    "true"
                )
    except EnvironmentError as e:
        report_processor.report(
            reports.file_io_error(
                file_type_codes.PCSD_ENVIRONMENT_CONFIG,
                RawFileError.ACTION_READ,
                format_environment_error(e),
                file_path=settings.pcsd_config,
            )
        )
    return False

def _verify_corosync_conf(corosync_conf_facade):
    # This is done in pcs.lib.env.LibraryEnvironment.push_corosync_conf
    # usually. But there are special cases here which use custom corosync.conf
    # pushing so the check must be done individually.
    bad_sections, bad_attr_names, bad_attr_values = (
        config_parser.verify_section(corosync_conf_facade.config)
    )
    if bad_sections or bad_attr_names or bad_attr_values:
        raise LibraryError(
            reports.corosync_config_cannot_save_invalid_names_values(
                bad_sections, bad_attr_names, bad_attr_values
            )
        )

def remove_nodes(env, node_list, force_flags=None):
    # pylint: disable=too-many-locals, too-many-branches, too-many-statements
    """
    Remove nodes from a cluster.

    env LibraryEnvironment
    node_list iterable -- names of nodes to remove
    force_flags list -- list of flags codes
    """
    _ensure_live_env(env) # raises if env is not live

    if force_flags is None:
        force_flags = []
    force_quorum_loss = report_codes.FORCE in force_flags
    skip_offline = report_codes.SKIP_OFFLINE_NODES in force_flags

    report_processor = SimpleReportProcessor(env.report_processor)
    target_factory = env.get_node_target_factory()
    corosync_conf = env.get_corosync_conf()

    # validations

    cluster_nodes_names, report_list = get_existing_nodes_names(
        corosync_conf,
        # Pcs is unable to communicate with nodes missing names. It cannot send
        # new corosync.conf to them. That might break the cluster. Hence we
        # error out.
        error_on_missing_name=True
    )
    report_processor.report_list(report_list)

    report_processor.report_list(config_validators.remove_nodes(
        node_list,
        corosync_conf.get_nodes(),
        corosync_conf.get_quorum_device_settings(),
    ))
    if report_processor.has_errors:
        # If there is an error, there is usually not much sense in doing other
        # validations:
        # - if there would be no node left in the cluster, it's pointless
        #   to check for quorum loss or if at least one remaining node is online
        # - if only one node is being removed and it doesn't exist, it's again
        #   pointless to check for other issues
        raise LibraryError()

    target_report_list, cluster_nodes_target_list = (
        target_factory.get_target_list_with_reports(
            cluster_nodes_names,
            skip_non_existing=skip_offline,
        )
    )
    known_nodes = {target.label for target in cluster_nodes_target_list}
    unknown_nodes = {
        name for name in cluster_nodes_names if name not in known_nodes
    }
    report_processor.report_list(target_report_list)

    com_cmd = GetOnlineTargets(
        report_processor, ignore_offline_targets=skip_offline,
    )
    com_cmd.set_targets(cluster_nodes_target_list)
    online_target_list = run_com(env.get_node_communicator(), com_cmd)
    offline_target_list = [
        target for target in cluster_nodes_target_list
        if target not in online_target_list
    ]
    staying_online_target_list = [
        target for target in online_target_list if target.label not in node_list
    ]
    targets_to_remove = [
        target for target in cluster_nodes_target_list
        if target.label in node_list
    ]
    if not staying_online_target_list:
        report_processor.report(
            reports.unable_to_connect_to_any_remaining_node()
        )
        # If no remaining node is online, there is no point in checking quorum
        # loss or anything as we would just get errors.
        raise LibraryError()

    if skip_offline:
        staying_offline_nodes = (
            [
                target.label for target in offline_target_list
                if target.label not in node_list
            ]
            +
            [name for name in unknown_nodes if name not in node_list]
        )
        if staying_offline_nodes:
            report_processor.report(
                reports.unable_to_connect_to_all_remaining_node(
                    staying_offline_nodes
                )
            )

    atb_has_to_be_enabled = sbd.atb_has_to_be_enabled(
        env.cmd_runner(), corosync_conf, -len(node_list)
    )
    if atb_has_to_be_enabled:
        report_processor.report(
            reports.corosync_quorum_atb_will_be_enabled_due_to_sbd()
        )
        com_cmd = CheckCorosyncOffline(
            report_processor, allow_skip_offline=False,
        )
        com_cmd.set_targets(staying_online_target_list)
        run_com(env.get_node_communicator(), com_cmd)
    else:
        # Check if removing the nodes would cause quorum loss. We ask the nodes
        # to be removed for their view of quorum. If they are all stopped or
        # not in a quorate partition, their removal cannot cause quorum loss.
        # That's why we ask them and not the remaining nodes.
        # example: 5-node cluster, 3 online nodes, removing one online node,
        # results in 4-node cluster with 2 online nodes => quorum lost
        # Check quorum loss only if ATB does not need to be enabled. If it is
        # required, cluster has to be turned off and therefore it loses quorum.
        forceable_report_creator = reports.get_problem_creator(
            report_codes.FORCE_QUORUM_LOSS, force_quorum_loss
        )
        com_cmd = cluster.GetQuorumStatus(report_processor)
        com_cmd.set_targets(targets_to_remove)
        failures, quorum_status = run_com(env.get_node_communicator(), com_cmd)
        if quorum_status:
            if quorum_status.stopping_nodes_cause_quorum_loss(node_list):
                report_processor.report(
                    forceable_report_creator(
                        reports.corosync_quorum_will_be_lost
                    )
                )
        elif failures or not targets_to_remove:
            report_processor.report(
                forceable_report_creator(
                    reports.corosync_quorum_loss_unable_to_check,
                )
            )

    if report_processor.has_errors:
        raise LibraryError()

    # validations done

    unknown_to_remove = [name for name in unknown_nodes if name in node_list]
    if unknown_to_remove:
        report_processor.report(
            reports.nodes_to_remove_unreachable(unknown_to_remove)
        )
    if targets_to_remove:
        com_cmd = cluster.DestroyWarnOnFailure(report_processor)
        com_cmd.set_targets(targets_to_remove)
        run_and_raise(env.get_node_communicator(), com_cmd)

    corosync_conf.remove_nodes(node_list)
    if atb_has_to_be_enabled:
        corosync_conf.set_quorum_options(dict(auto_tie_breaker="1"))

    _verify_corosync_conf(corosync_conf) # raises if corosync not valid
    com_cmd = DistributeCorosyncConf(
        env.report_processor,
        corosync_conf.config.export(),
        allow_skip_offline=False,
    )
    com_cmd.set_targets(staying_online_target_list)
    run_and_raise(env.get_node_communicator(), com_cmd)

    com_cmd = ReloadCorosyncConf(env.report_processor)
    com_cmd.set_targets(staying_online_target_list)
    run_and_raise(env.get_node_communicator(), com_cmd)

    # try to remove nodes from pcmk using crm_node -R <node> --force and if not
    # successful remove it directly from CIB file on all nodes in parallel
    com_cmd = RemoveNodesFromCib(env.report_processor, node_list)
    com_cmd.set_targets(staying_online_target_list)
    run_and_raise(env.get_node_communicator(), com_cmd)


def remove_nodes_from_cib(env, node_list):
    """
    Remove specified nodes from CIB. When pcmk is running 'crm_node -R <node>'
    will be used. Otherwise nodes will be removed directly from CIB file.

    env LibraryEnvironment
    node_list iterable -- names of nodes to remove
    """
    # TODO: more advanced error handling
    if not env.is_cib_live:
        raise LibraryError(reports.live_environment_required(["CIB"]))

    if is_service_running(env.cmd_runner(), "pacemaker"):
        for node in node_list:
            # this may raise a LibraryError
            # NOTE: crm_node cannot remove multiple nodes at once
            remove_node(env.cmd_runner(), node)
        return

    # TODO: We need to remove nodes from the CIB file. We don't want to do it
    # using environment as this is a special case in which we have to edit CIB
    # file directly.
    for node in node_list:
        stdout, stderr, retval = env.cmd_runner().run(
            [
                settings.cibadmin,
                "--delete-all",
                "--force",
                f"--xpath=/cib/configuration/nodes/node[@uname='{node}']",
            ],
            env_extend={"CIB_file": os.path.join(settings.cib_dir, "cib.xml")}
        )
        if retval != 0:
            raise LibraryError(
                reports.node_remove_in_pacemaker_failed(
                    [node],
                    reason=join_multilines([stderr, stdout])
                )
            )

def add_link(env, node_addr_map, link_options=None, force_flags=None):
    """
    Add a corosync link to a cluster

    env LibraryEnvironment
    dict node_addr_map -- key: node name, value: node address for the link
    dict link_options -- link options
    force_flags list -- list of flags codes
    """
    _ensure_live_env(env) # raises if env is not live

    link_options = link_options or dict()
    force_flags = force_flags or set()
    force = report_codes.FORCE in force_flags
    skip_offline = report_codes.SKIP_OFFLINE_NODES in force_flags

    report_processor = SimpleReportProcessor(env.report_processor)
    corosync_conf = env.get_corosync_conf()

    # validations

    dummy_cluster_nodes_names, nodes_report_list = get_existing_nodes_names(
        corosync_conf,
        # New link addresses are assigned to nodes based on node names. If
        # there are nodes with no names, we cannot assign them new link
        # addresses. This is a no-go situation.
        error_on_missing_name=True
    )
    report_processor.report_list(nodes_report_list)

    try:
        cib = env.get_cib()
        cib_nodes = get_remote_nodes(cib) + get_guest_nodes(cib)
    except LibraryError:
        cib_nodes = []
        report_processor.report(
            reports.get_problem_creator(
                report_codes.FORCE_LOAD_NODES_FROM_CIB,
                force
            )(
                reports.cib_load_error_get_nodes_for_validation
            )
        )

    report_processor.report_list(config_validators.add_link(
        node_addr_map,
        link_options,
        corosync_conf.get_nodes(),
        cib_nodes,
        corosync_conf.get_used_linknumber_list(),
        corosync_conf.get_transport(),
        corosync_conf.get_ip_version(),
        force_unresolvable=force
    ))

    if report_processor.has_errors:
        raise LibraryError()

    # validations done

    corosync_conf.add_link(node_addr_map, link_options)
    env.push_corosync_conf(corosync_conf, skip_offline)

def remove_links(env, linknumber_list, force_flags=None):
    """
    Remove corosync links from a cluster

    env LibraryEnvironment
    iterable linknumber_list -- linknumbers (as strings) of links to be removed
    force_flags list -- list of flags codes
    """
    # TODO library interface should make sure linknumber_list is an iterable of
    # strings. The layer in which the check should be done does not exist yet.
    _ensure_live_env(env) # raises if env is not live

    force_flags = force_flags or set()
    skip_offline = report_codes.SKIP_OFFLINE_NODES in force_flags

    report_processor = SimpleReportProcessor(env.report_processor)
    corosync_conf = env.get_corosync_conf()

    # validations

    report_processor.report_list(config_validators.remove_links(
        linknumber_list,
        corosync_conf.get_used_linknumber_list(),
        corosync_conf.get_transport()
    ))

    if report_processor.has_errors:
        raise LibraryError()

    # validations done

    corosync_conf.remove_links(linknumber_list)

    env.push_corosync_conf(corosync_conf, skip_offline)

def update_link(
    env, linknumber, node_addr_map=None, link_options=None, force_flags=None
):
    """
    Change an existing corosync link

    env LibraryEnvironment
    string linknumber -- the link to be changed
    dict node_addr_map -- key: node name, value: node address for the link
    dict link_options -- link options
    force_flags list -- list of flags codes
    """
    _ensure_live_env(env) # raises if env is not live

    node_addr_map = node_addr_map or dict()
    link_options = link_options or dict()
    force_flags = force_flags or set()
    force = report_codes.FORCE in force_flags
    skip_offline = report_codes.SKIP_OFFLINE_NODES in force_flags

    report_processor = SimpleReportProcessor(env.report_processor)
    corosync_conf = env.get_corosync_conf()

    # validations

    dummy_cluster_nodes_names, nodes_report_list = get_existing_nodes_names(
        corosync_conf,
        # Pcs is unable to communicate with nodes missing names. It cannot send
        # new corosync.conf to them. That might break the cluster. Hence we
        # error out.
        # This check is done later as well, when sending corosync.conf to
        # nodes. But we need node names to be present so we can set new
        # addresses to them. We may as well do the check right now.
        error_on_missing_name=True
    )
    report_processor.report_list(nodes_report_list)

    report_processor.report_list(config_validators.update_link(
        linknumber,
        node_addr_map,
        link_options,
        corosync_conf.get_links_options().get(linknumber, {}),
        corosync_conf.get_nodes(),
        # cluster must be stopped for updating a link and then we cannot get
        # nodes from CIB
        [],
        corosync_conf.get_used_linknumber_list(),
        corosync_conf.get_transport(),
        corosync_conf.get_ip_version(),
        force_unresolvable=force
    ))

    if report_processor.has_errors:
        raise LibraryError()

    # validations done

    corosync_conf.update_link(linknumber, node_addr_map, link_options)
    env.push_corosync_conf(corosync_conf, skip_offline)
